# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import importlib
import inspect
import logging
from operator import itemgetter
from pkgutil import iter_modules
import sys

logger = logging.getLogger(__name__)


def get_trackers_list(include_disabled=False, order_by=('name',)):
    trackers = []
    for module in get_trackers_modules():
        for _name, obj in dict(inspect.getmembers(module)).items():
            if not hasattr(obj, 'id') or not hasattr(obj, 'name'):
                continue
            if NotImplemented in (obj.id, obj.name):
                continue

            if not include_disabled and obj.status == 'disabled':
                continue

            if inspect.isclass(obj):
                trackers.append(dict(
                    id=obj.id,
                    name=obj.name,
                    module=module,
                    class_name=obj.id.capitalize(),
                ))

    return sorted(trackers, key=itemgetter(*order_by))


def get_trackers_modules():
    def import_modules(namespace, modules, modules_names):
        count = 0
        for _finder, module_name, ispkg in iter_namespace(namespace):
            if module_name in modules_names or not ispkg:
                continue

            module = importlib.import_module(module_name)
            modules.append(module)
            modules_names.append(module_name)
            count += 1

        if count > 0:
            logger.info('Import {0} trackers modules'.format(count))

        return count

    def iter_namespace(ns_pkg):
        # Specifying the second argument (prefix) to iter_modules makes the
        # returned name an absolute name instead of a relative one. This allows
        # import_module to work without having to do additional modification to
        # the name.
        return iter_modules(ns_pkg.__path__, ns_pkg.__name__ + '.')

    modules = []
    modules_names = []
    for finder in sys.meta_path:
        import komikku.trackers

        import_modules(komikku.trackers, modules, modules_names)

    return modules



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number


class Zeroscans(Server):
    id = 'zeroscans'
    name = 'Zero Scans'
    lang = 'en'

    base_url = 'https://zscans.com'
    logo_url = base_url + '/favicon.ico'
    api_url = base_url + '/swordflake'
    api_latest_updates_url = api_url + '/new-chapters'
    api_search_url = api_url + '/comics'
    api_manga_url = api_url + '/comic/{0}'
    api_chapter_url = api_url + '/comic/{0}/chapters/{1}'
    api_chapters_url = api_url + '/comic/{0}/chapters?sort=desc&page={1}'
    manga_url = base_url + '/comics/{0}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns comic data using API

        Initial data should contain at least comic's slug (provided by search)
        """
        r = self.session_get(self.api_manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        resp_data = r.json()['data']

        initial_data.pop('view_count', None)
        data = initial_data.copy()
        data.update(dict(
            name=resp_data['name'],
            authors=[],     # Not available
            scanlators=[],  # Not available
            genres=[genre['name'] for genre in resp_data['genres']],
            status=None,
            synopsis=resp_data['summary'],
            chapters=[],
            server_id=self.id,
            cover=resp_data['cover']['full'],
        ))

        # Status
        status = resp_data['statuses'][0]['slug']
        if status in ('new', 'ongoing'):
            data['status'] = 'ongoing'
        elif status == 'completed':
            data['status'] = 'complete'
        elif status == 'dropped':
            data['status'] = 'suspended'
        elif status == 'hiatus':
            data['status'] = 'hiatus'

        # Chapters
        id_ = resp_data['id']
        page = 1
        next_page = True
        chapters = []
        while next_page:
            r = self.session_get(self.api_chapters_url.format(id_, page))
            if r.status_code != 200:
                chapters = []
                break

            resp_data = r.json()
            if not resp_data['success']:
                chapters = []
                break

            resp_data = resp_data['data']
            for chapter in resp_data['data']:
                chapters.append(dict(
                    slug=str(chapter['id']),
                    title=f'#{chapter["name"]}',
                    num=chapter['name'] if is_number(chapter['name']) else None,
                    date=convert_date_string(chapter['created_at']),
                ))

            next_page = resp_data['next_page_url'] is not None
            page += 1

        data['chapters'] = list(reversed(chapters))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns comic chapter data using API

        Currently, only pages are expected.
        """
        r = self.session_get(self.api_chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        resp_data = r.json()
        if not resp_data['success']:
            return None

        data = dict(
            pages=[],
        )
        for url in resp_data['data']['chapter']['good_quality']:
            data['pages'].append(dict(
                slug=None,
                image=url,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns comic absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        r = self.session_get(self.api_latest_updates_url)
        if r.status_code != 200:
            return None

        data = r.json()

        result = []
        slugs = []
        for item in data['all']:
            if item['slug'] in slugs:
                continue

            result.append(dict(
                name=item['name'],
                slug=item['slug'],
                cover=item['cover'].get('vertical') or item['cover'].get('horizontal'),
            ))
            slugs.append(item['slug'])

        return result

    def get_most_populars(self):
        return self.search('', True)

    def search(self, term, most_populars=False):
        r = self.session_get(self.api_search_url)
        if r.status_code != 200:
            return None

        data = r.json()
        if not data['success']:
            return None

        data = data['data']

        result = []
        for item in data['comics']:
            if not most_populars and term.lower() not in item['name'].lower():
                continue

            result.append(dict(
                name=item['name'],
                slug=item['slug'],
                cover=item['cover']['horizontal'],
                nb_chapters=item['chapter_count'],
                view_count=item['view_count'],
            ))

        if most_populars:
            result = sorted(result, key=lambda i: i['view_count'], reverse=True)

        return result



# SPDX-FileCopyrightText: 2020-2024 GrownNed
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: GrownNed <grownned@gmail.com>
# Author: valos <vfebvre@easter-eggs.com>

import uuid

try:
    # For some reasons, under flatpak sandbox, HTTP requests return 403 errors!
    # Only solution found, use curl_cffi (JA3/TLS and HTTP2 fingerprints impersonation) in place of requests
    # What's wrong with sandbox?
    from curl_cffi import requests
except Exception:
    # Server will be disabled
    requests = None

from komikku.consts import REQUESTS_TIMEOUT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number


class Mangalib(Server):
    id = 'mangalib'
    name = 'MangaLIB'
    lang = 'ru'

    base_url = 'https://mangalib.org'
    logo_url = base_url + '/static/images/logo/ml/favicon.ico?529e41ba742f20a52986d0dea3e6ab1b'
    manga_url = base_url + '/ru/manga/{0}'
    api_base_url = 'https://api.cdnlibs.org/api'
    api_search_url = api_base_url + '/manga'
    api_manga_url = api_base_url + '/manga/{0}'
    api_chapters_url = api_base_url + '/manga/{0}/chapters'
    api_chapter_url = api_base_url + '/manga/{0}/chapter'
    image_base_url = 'https://img3.mixlib.me/'  # beware, an additional slash is required

    site_id = 1

    api_headers = {
        'Accept': '*/*',
        'Accept-Language': 'en-US;q=0.5,en;q=0.3',
        'Client-Time-Zone': 'Atlantic/Reykjavik',
        'Content-Type': 'application/json',
        'Referer': f'{base_url}/',
        'Site-Id': site_id,
    }

    def __init__(self):
        if self.session is None and requests is not None:
            self.session = requests.Session(
                allow_redirects=True,
                impersonate='chrome',
                timeout=(REQUESTS_TIMEOUT, REQUESTS_TIMEOUT * 2),
            )

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        params = {
            'fields[]': ['background', 'eng_name', 'otherNames', 'summary', 'releaseDate', 'type_id', 'caution', 'views', 'close_view', 'rate_avg', 'rate', 'genres', 'tags', 'teams', 'user', 'franchise', 'authors', 'publisher', 'userRating', 'moderated', 'metadata', 'metadata.count', 'metadata.close_comments', 'manga_status_id', 'chap_count', 'status_id', 'artists', 'format']
        }
        r = self.session_get(
            self.api_manga_url.format(initial_data['slug']),
            params=params,
            headers=self.api_headers
        )
        if r.status_code != 200:
            return None

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        resp_data = r.json()['data']

        data['name'] = resp_data['rus_name']
        data['cover'] = resp_data['cover']['thumbnail']

        # Details
        for genre in resp_data['genres']:
            data['genres'].append(genre['name'])
        data['genres'].append(resp_data['type']['label'])

        for author in resp_data['authors']:
            data['authors'].append(author['name'])
        for artist in resp_data['artists']:
            author = artist['name']
            if author in data['authors']:
                continue
            data['authors'].append(author)

        for team in resp_data['teams']:
            data['scanlators'].append(team['name'])

        status = resp_data['status']['label']
        if status == 'Онгоинг':
            data['status'] = 'ongoing'
        elif status == 'Завершён':
            data['status'] = 'complete'
        elif status == 'Приостановлен':
            data['status'] = 'suspended'
        elif status == 'Выпуск прекращён':
            data['status'] = 'suspended'

        data['synopsis'] = resp_data['summary']

        # Chapters
        r = self.session_get(
            self.api_chapters_url.format(initial_data['slug']),
            headers=self.api_headers
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()['data']

        for chapter in resp_data:
            chapter_id = chapter['id']
            date = None
            scanlators = []
            for branch in chapter['branches']:
                if branch['id'] != chapter_id:
                    continue

                date = branch['created_at'][:10]
                for team in branch['teams']:
                    scanlators.append(team['name'])

            data['chapters'].append(dict(
                slug=f'v{chapter["volume"]}/c{chapter["number"]}',
                title=f'Том {chapter["volume"]} Глава {chapter["number"]} - {chapter["name"]}',
                num=chapter['number'] if is_number(chapter['number']) else None,
                num_volume=chapter['volume'] if is_number(chapter['volume']) else None,
                date=convert_date_string(date, format='%Y-%m-%d') if date else None,
                scanlators=scanlators if scanlators else None,
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        volume, number = chapter_slug.split('/')
        r = self.session_get(
            self.api_chapter_url.format(manga_slug),
            params={
                'number': number[1:],
                'volume': volume[1:],
            },
            headers=self.api_headers
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()['data']

        data = dict(
            pages=[],
        )
        for page in resp_data['pages']:
            data['pages'].append(dict(
                slug=None,
                image=page['url'][1:],
                index=page['slug'],
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            self.image_base_url + page['image'],
            headers={
                'Referer': f'{self.base_url}/'
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0:03d}.{1}'.format(page['index'], mime_type.split('/')[-1]),
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        """
        Returns latest updates
        """
        return self.search('', orderby='latest')

    def get_most_populars(self):
        """
        Returns best noted manga list
        """
        return self.search('', orderby='populars')

    def search(self, term, orderby=None):
        params = {
            'fields[]': ['rate', 'rate_avg', 'userBookmark'],
            'seed': uuid.uuid1().hex,
            'site_id[]': self.site_id,
        }
        r = self.session_get(
            self.api_search_url,
            params=params,
            headers=self.api_headers,
        )
        seed = r.json()['meta']['seed']

        if orderby == 'latest':
            params.update({
                'sort_by': 'last_chapter_at',
                'seed': seed,
            })
        elif orderby == 'populars':
            params.update({
                'sort_by': 'views',
                'seed': seed,
            })
        else:
            params.update({
                'q': term,
                'seed': seed,
            })

        r = self.session_get(
            self.api_search_url,
            params=params,
            headers=self.api_headers,
        )
        if r.status_code != 200:
            return None

        results = []
        for item in r.json()['data']:
            results.append(dict(
                name=item['name'],
                slug=item['slug_url'],
                cover=item['cover']['thumbnail'],
            ))

        return results


# NSFW
class Hentailib(Mangalib):
    id = 'hentailib:mangalib'
    name = 'HentaiLIB'
    lang = 'ru'
    is_nsfw_only = True

    # Unfortunately, user authentication (via social networks) is required to get chapter data
    status = 'disabled'

    base_url = 'https://hentailib.me'
    manga_url = base_url + '/ru/manga/{0}'
    api_base_url = 'https://hapi.hentaicdn.org/api'
    api_search_url = api_base_url + '/manga'
    api_manga_url = api_base_url + '/manga/{0}'
    api_chapters_url = api_base_url + '/manga/{0}/chapters'
    api_chapter_url = api_base_url + '/manga/{0}/chapter'

    site_id = 4

    api_headers = {
        'Accept': '*/*',
        'Accept-Language': 'en-US;q=0.5,en;q=0.3',
        'Client-Time-Zone': 'Atlantic/Reykjavik',
        'Content-Type': 'application/json',
        'Referer': f'{base_url}/',
        'Site-Id': f'{site_id}',
    }



# SPDX-FileCopyrightText: 2020-2024 Liliana Prikler
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Liliana Prikler <liliana.prikler@gmail.com>

from bs4 import BeautifulSoup
import json

from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.webview import CompleteChallenge

IMAGES_EXTS = {'g': 'gif', 'j': 'jpg', 'p': 'png', 'w': 'webp'}

# Mirrors
# https://nhentai.to
# https://nhentai.xxx


class Nhentai(Server):
    id = 'nhentai'
    name = 'NHentai'
    lang = 'en'
    lang_code = 'english'
    is_nsfw_only = True

    has_cf = True

    base_url = 'https://nhentai.net'
    search_url = base_url + '/search'
    manga_url = base_url + '/g/{0}'
    page_image_url = 'https://i.nhentai.net/galleries/{0}/{1}'

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        data = initial_data.copy()
        data.update({
            'authors': [],
            'scanlators': [],  # Not available
            'genres': [],
            'status': None,    # Not available
            'synopsis': None,  # Not available
            'chapters': [],
            'server_id': self.id,
        })

        soup = BeautifulSoup(r.text, 'lxml')

        data['name'] = soup.find('meta', property='og:title')['content']
        data['cover'] = 'https:' + soup.find('meta', property='og:image')['content']

        # Genres & Artists + chapter date
        chapter_date = None
        info = soup.select_one('#info')
        for tag_container in info.select('#tags .tag-container'):
            category = tag_container.text.split(':')[0].strip()

            if category == 'Uploaded':
                if time_element := tag_container.select_one('time'):
                    chapter_date = time_element.get('datetime').split('T')[0]

            for tag in tag_container.select('.tag'):
                clean_tag = tag.select_one('span.name').text.strip()
                if category in ['Artists', 'Groups', ]:
                    data['authors'].append(clean_tag)
                if category in ['Tags', ]:
                    data['genres'].append(clean_tag)

        # Single chapter
        data['chapters'].append({
            'slug': data['cover'].rstrip('/').split('/')[-2],
            'title': data['name'],
            'date': convert_date_string(chapter_date, '%Y-%m-%d') if chapter_date else None,
        })

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.manga_url.format(manga_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        pages = []
        for script_element in soup.find_all('script'):
            script = script_element.string
            if not script or not script.strip().startswith('window._gallery'):
                continue

            info = json.loads(script.strip().split('\n')[0][30:-3].replace('\\u0022', '"').replace('\\u005C', '\\'))
            if not info.get('images') or not info['images'].get('pages'):
                break

            for index, page in enumerate(info['images']['pages']):
                num = index + 1
                extension = IMAGES_EXTS[page['t']]
                page = {
                    'image': None,
                    'slug': f'{num}.{extension}',
                }
                pages.append(page)

        return {'pages': pages}

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        assert chapter_slug is not None
        r = self.session_get(self.page_image_url.format(chapter_slug, page['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': page['slug'],
        }

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def _search_common(self, params):
        r = self.session_get(self.search_url, params=params)

        if r.status_code == 200:
            try:
                results = []
                soup = BeautifulSoup(r.text, 'lxml')
                elements = soup.find_all('div', class_='gallery')

                for element in elements:
                    a_element = element.find('a', class_='cover')
                    caption_element = element.find('div', class_='caption')
                    results.append({
                        'slug': a_element.get('href').rstrip('/').split('/')[-1],
                        'name': caption_element.text.strip(),
                        'cover': 'https:' + a_element.img.get('data-src'),
                    })
            except Exception:
                return None
            else:
                return results

        return None

    @CompleteChallenge()
    def get_most_populars(self):
        """
        Returns most popular mangas (bayesian rating)
        """
        return self._search_common({'q': 'language:' + self.lang_code, 'sort': 'popular'})

    @CompleteChallenge()
    def search(self, term):
        term = term + ' language:' + self.lang_code
        return self._search_common({'q': term})


class Nhentai_chinese(Nhentai):
    id = 'nhentai_chinese'
    lang = 'zh_Hans'
    lang_code = 'chinese'


class Nhentai_japanese(Nhentai):
    id = 'nhentai_japanese'
    lang = 'ja'
    lang_code = 'japanese'



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from bs4 import BeautifulSoup

from komikku.servers.multi.heancms import HeanCMS


class Kenscans(HeanCMS):
    id = 'kenscans'
    name = 'Kenscans'
    lang = 'en'

    base_url = 'https://kencomics.com'
    logo_url = base_url + '/favicon-32x32.png'
    api_url = 'https://api.kencomics.com/api'

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Pages URLs are available in a <script> element
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for image in soup.select('.image-container img'):
            data['pages'].append(dict(
                slug=None,
                image=image.get('src'),
            ))

        return data



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Graphene
from gi.repository import Gsk
from gi.repository import Gtk
from gi.repository import Pango

from komikku.consts import COVER_HEIGHT
from komikku.consts import COVER_WIDTH
from komikku.consts import MISSING_IMG_RESOURCE_PATH
from komikku.models import Settings
from komikku.utils import CoverLoader


class LibraryFlowBoxChildBase:
    default_width = COVER_WIDTH
    default_height = COVER_HEIGHT
    margins = (3 + 3 + 3) * 2  # FlowBoxChild margin + FlowBoxChild border + FlowBoxChild child margin
    server_logo_size = 16

    cover_picture: Gtk.Picture
    name_label: Gtk.Label

    def __init__(self, parent, manga):
        self.parent = parent
        self.manga = manga
        self._filtered = False
        self._selected = False

        self.cover_picture.set_paintable(LibraryFlowBoxChildCoverPaintable(manga))

        # Logo widget (Gtk.Image or Adw.Avatar as fallback)
        if Settings.get_default().library_servers_logo:
            if self.manga.server.logo_path:
                self.logo = Gtk.Image.new_from_file(self.manga.server.logo_path)
                self.logo.set_pixel_size(self.server_logo_size)
            elif self.manga.server.id == 'local':
                self.logo = Adw.Avatar.new(self.server_logo_size, None, False)
                self.logo.set_icon_name('folder-symbolic')
            else:
                self.logo = Adw.Avatar.new(self.server_logo_size, self.manga.server.name, True)
        else:
            self.logo = None

    def draw_name(self):
        self.name_label.set_text(self.manga.name)

    def resize_cover(self, width, height):
        cover = self.cover_picture.get_paintable()
        if cover.width == width:
            return

        cover.resize(width, height)

    def update(self, manga=None):
        if manga:
            self.manga = manga
            self.draw_name()

        self.cover_picture.get_paintable().update(manga)


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/library_compact_flowbox_child.ui')
class LibraryCompactFlowBoxChild(Gtk.FlowBoxChild, LibraryFlowBoxChildBase):
    __gtype_name__ = 'LibraryCompactFlowBoxChild'

    overlay = Gtk.Template.Child('overlay')
    cover_picture = Gtk.Template.Child('cover_picture')
    name_label = Gtk.Template.Child('name_label')

    def __init__(self, parent, manga, width, height):
        Gtk.FlowBoxChild.__init__(self)
        LibraryFlowBoxChildBase.__init__(self, parent, manga)

        if self.logo:
            self.logo.props.margin_start = 6
            self.logo.props.margin_top = 6
            self.logo.props.halign = Gtk.Align.START
            self.logo.props.valign = Gtk.Align.START
            self.overlay.add_overlay(self.logo)

        self.draw_name()
        self.resize_cover(width, height)


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/library_flowbox_child.ui')
class LibraryFlowBoxChild(Gtk.FlowBoxChild, LibraryFlowBoxChildBase):
    __gtype_name__ = 'LibraryFlowBoxChild'

    grid = Gtk.Template.Child('grid')
    cover_picture = Gtk.Template.Child('cover_picture')
    name_label = Gtk.Template.Child('name_label')

    def __init__(self, parent, manga, width, height):
        Gtk.FlowBoxChild.__init__(self)
        LibraryFlowBoxChildBase.__init__(self, parent, manga)

        if self.logo:
            self.name_label.props.xalign = 0

            self.logo.props.halign = Gtk.Align.END
            self.logo.props.valign = Gtk.Align.CENTER

            self.grid.attach(self.logo, 1, 1, 1, 1)
        else:
            self.name_label.set_justify(Gtk.Justification.CENTER)

        self.draw_name()
        self.resize_cover(width, height)


class LibraryFlowBoxChildCoverPaintable(GObject.GObject, Gdk.Paintable):
    __gtype_name__ = 'LibraryFlowBoxChildCoverPaintable'

    badge_font_size = 9
    badge_layout = None
    badge_text_color = Gdk.RGBA(1, 1, 1, 1)
    rect = Graphene.Rect().alloc()
    rounded_rect = Gsk.RoundedRect()
    rounded_rect_size = Graphene.Size.init(Graphene.Size().alloc(), 8, 8)

    def __init__(self, manga):
        super().__init__()

        self.manga = manga
        self.cover_texture = None
        self.width = None
        self.height = None

        if LibraryFlowBoxChildCoverPaintable.badge_layout is None:
            pango_context = Gio.Application.get_default().window.get_pango_context()

            font = pango_context.get_font_description()
            font.set_size(self.badge_font_size * Pango.SCALE)
            font.set_weight(Pango.Weight.HEAVY)

            LibraryFlowBoxChildCoverPaintable.badge_layout = Pango.Layout(pango_context)
            LibraryFlowBoxChildCoverPaintable.badge_layout.set_font_description(font)

        self.__get_badges_values()
        self.__create_cover_texture()

    def __create_cover_texture(self):
        if self.manga.cover_fs_path is None:
            paintable = CoverLoader.new_from_resource(MISSING_IMG_RESOURCE_PATH, COVER_WIDTH, None)
        else:
            paintable = CoverLoader.new_from_file(self.manga.cover_fs_path, COVER_WIDTH, None, True)
            if paintable is None:
                paintable = CoverLoader.new_from_resource(MISSING_IMG_RESOURCE_PATH, COVER_WIDTH, None)

        self.cover_texture = Gdk.Texture.new_for_pixbuf(paintable.pixbuf) if paintable.pixbuf else paintable.texture

    def __get_badges_values(self):
        badges = Settings.get_default().library_badges
        self.nb_unread_chapters = self.manga.nb_unread_chapters if 'unread-chapters' in badges else None
        self.nb_downloaded_chapters = self.manga.nb_downloaded_chapters if 'downloaded-chapters' in badges else None
        self.nb_recent_chapters = self.manga.nb_recent_chapters if 'recent-chapters' in badges else None

    def do_get_intrinsic_height(self):
        return self.height

    def do_get_intrinsic_width(self):
        return self.width

    def do_snapshot(self, snapshot, width, height):
        self.rect.init(0, 0, width, height)

        # Draw cover (rounded)
        self.rounded_rect.init(self.rect, self.rounded_rect_size, self.rounded_rect_size, self.rounded_rect_size, self.rounded_rect_size)
        snapshot.push_rounded_clip(self.rounded_rect)
        snapshot.append_texture(self.cover_texture, self.rect)
        snapshot.pop()  # remove the clip

        # Draw badges (top right corner)
        spacing = 5  # with top border, right border and between badges
        x = width

        def draw_badge(value, color):
            nonlocal x

            if not value:
                return

            self.badge_layout.set_text(str(value))
            extent = self.badge_layout.get_pixel_extents()[1]
            w = extent.width + 2 * 7
            h = extent.height + 2 * 1

            # Draw rounded rectangle (pill)
            x = x - spacing - w
            y = spacing

            bg_color = Gdk.RGBA()
            bg_color.parse(color)

            rect = self.rect.init(x, y, w, h)
            self.rounded_rect.init_from_rect(self.rect, radius=90)

            snapshot.push_rounded_clip(self.rounded_rect)
            snapshot.append_color(bg_color, rect)
            snapshot.pop()  # remove the clip

            # Draw number
            snapshot.save()
            snapshot.translate(Graphene.Point().init(x + 7, y + 1))
            snapshot.append_layout(self.badge_layout, self.badge_text_color)
            snapshot.restore()

        draw_badge(self.nb_unread_chapters, '#62a0ea')      # @blue_2
        draw_badge(self.nb_downloaded_chapters, '#f68276')
        draw_badge(self.nb_recent_chapters, '#33d17a')      # @green_3

    def resize(self, width, height):
        self.width = width
        self.height = height

        self.invalidate_size()

    def update(self, manga=None):
        if manga:
            self.manga = manga
            self.__create_cover_texture()

        self.__get_badges_values()

        self.invalidate_contents()



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import json
import re

from bs4 import BeautifulSoup

from komikku.servers.multi.madara import Madara2
from komikku.utils import get_buffer_mime_type

# 2025/01 fusion of Sinensis and Cerise => SCtoon (Peachscan)
# 2025/09 SCtoon => Lovers Toon (Madara)


class Sinensistoon(Madara2):
    id = 'sinensistoon'
    name = 'Lovers Toon (Sinensis/Cerise)'
    lang = 'pt_BR'
    status = 'disabled'  # Dead 05/2026 => loverstoon.net

    date_format = '%d.%m.%Y'

    base_url = 'https://loverstoon.com'
    logo_url = base_url + '/wp-content/uploads/2025/09/cropped-faviliocon-32x32.png'
    chapters_url = base_url + '/manga/{0}/ajax/chapters/'

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )

        found = False
        if script_element := soup.select_one('.page-break script'):
            script = script_element.string
            if script:
                re_pages = r'content:\s+(\[[\s\S*]+\])'
                if matches := re.search(re_pages, script):
                    found = True
                    pages = json.loads(matches.group(1))
                    for index, url in enumerate(pages):
                        data['pages'].append(dict(
                            slug=None,
                            image=url,
                            index=index + 1,
                        ))

        if not found:
            return None

        return data



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.foolslide import FoOlSlide

#
# BEWARE: Kirei Cake server is disabled
# Dead since 11/2022
# Migrate as MangaDex group: https://mangadex.org/group/5fa94491-6f6f-4d53-b8ff-4a4967ac40b5/kirei-cake?tab=feed
#


class Kireicake(FoOlSlide):
    id = 'kireicake'
    name = 'Kirei Cake'
    lang = 'en'
    status = 'disabled'

    base_url = 'https://reader.kireicake.com'
    search_url = base_url + '/search'
    mangas_url = base_url + '/directory'
    manga_url = base_url + '/series/{0}'
    chapter_url = base_url + '/read/{0}/en/{1}/page/1'



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import logging

import gi

gi.require_version('Gtk', '4.0')

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Graphene
from gi.repository import Gsk
from gi.repository import Gtk

from komikku.models import Settings

logger = logging.getLogger('komikku')

PRELOAD = 5  # in widget height unit
ZOOM_FACTOR_SCROLL_WHEEL = 1.3


class KInfiniteCanvas(Gtk.Widget, Gtk.Scrollable):
    __gtype_name__ = 'KInfiniteCanvas'
    __gsignals__ = {
        'controls-zone-clicked': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'keyboard-navigation': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'offlimit': (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        'page-requested': (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self, pager):
        super().__init__()

        self.pager = pager
        self.settings = Settings.get_default()

        self.__interactive = True
        self.__hadj = None
        self.__vadj = None
        self.zoom = 1

        self.vadjustment_value = None
        self.vadjustment_value_changed_handler_id = None

        self.scroll_adjusting_delta = 0
        self.scroll_direction = None
        self.scroll_drag_offset = 0

        self.is_scroll_adjusting = False
        self.is_scroll_by_increment = False
        self.is_scroll_decelerating = False

        self.current_page_top = None
        self.current_page_bottom = None

        self.prev_width = None
        self.canvas_height = 0

        self.set_overflow(Gtk.Overflow.HIDDEN)

        self.add_controllers()

    @property
    def canvas_width(self):
        return self.widget_width

    @GObject.Property(type=Gtk.Adjustment)
    def hadjustment(self):
        return self.__hadj

    @hadjustment.setter
    def hadjustment(self, adj):
        self.__hadj = adj

    @GObject.Property(type=Gtk.ScrollablePolicy, default=Gtk.ScrollablePolicy.MINIMUM)
    def hscroll_policy(self):
        return Gtk.ScrollablePolicy.MINIMUM

    @GObject.Property(type=bool, default=True)
    def interactive(self):
        return self.__interactive

    @interactive.setter
    def interactive(self, value):
        self.__interactive = value

        if value:
            # Restore value
            if self.vadjustment_value is not None:
                self.vadjustment.props.value = self.vadjustment_value
            self.connect_signals()

            self.gesture_drag.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
            self.gesture_click.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
            self.controller_scroll.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        else:
            # Save value
            self.vadjustment_value = self.vadjustment.props.value
            self.disconnect_signals()

            self.gesture_drag.set_propagation_phase(Gtk.PropagationPhase.NONE)
            self.gesture_click.set_propagation_phase(Gtk.PropagationPhase.NONE)
            self.controller_scroll.set_propagation_phase(Gtk.PropagationPhase.NONE)

    @property
    def max_vadjustment_value(self):
        return max(self.canvas_height - self.widget_height, 0)

    @property
    def pages(self):
        pages = []
        page = self.get_first_child()
        while page:
            pages.append(page)
            page = page.get_next_sibling()

        return pages

    @GObject.Property(type=Gtk.Adjustment)
    def vadjustment(self):
        return self.__vadj or Gtk.Adjustment()

    @vadjustment.setter
    def vadjustment(self, adj):
        if not adj:
            return

        self.__vadj = adj
        self.connect_signals()

    @GObject.Property(type=Gtk.ScrollablePolicy, default=Gtk.ScrollablePolicy.MINIMUM)
    def vscroll_policy(self):
        return Gtk.ScrollablePolicy.MINIMUM

    @property
    def widget_height(self):
        return self.get_height()

    @property
    def widget_width(self):
        return self.get_width()

    def add_controllers(self):
        # Scroll controller
        self.controller_scroll = Gtk.EventControllerScroll.new(
            Gtk.EventControllerScrollFlags.VERTICAL | Gtk.EventControllerScrollFlags.KINETIC
        )
        self.controller_scroll.connect('decelerate', self.on_scroll_decelerate)
        self.controller_scroll.connect('scroll', self.on_scroll)
        self.add_controller(self.controller_scroll)

        # Gesture drag controller
        self.gesture_drag = Gtk.GestureDrag.new()
        self.gesture_drag.connect('drag-begin', self.on_gesture_drag_begin)
        self.gesture_drag.connect('drag-end', self.on_gesture_drag_end)
        self.gesture_drag.connect('drag-update', self.on_gesture_drag_update)
        self.pager.add_controller(self.gesture_drag)

        # Gesture click controller: Navigation layout
        self.gesture_click = Gtk.GestureClick.new()
        self.gesture_click.set_button(1)
        self.gesture_click.connect('released', self.on_gesture_click_released)
        self.add_controller(self.gesture_click)

    def add_or_remove_page(self):
        if not self.get_first_child() or self.is_scroll_decelerating or self.is_scroll_adjusting or self.is_scroll_by_increment:
            return

        first_page = self.get_first_child()
        last_page = self.get_last_child()
        if self.scroll_direction in (Gtk.DirectionType.DOWN, None):
            self.scroll_direction = Gtk.DirectionType.DOWN

            if last_page.loadable and last_page._ic_position + last_page._ic_height < self.widget_height * (PRELOAD + 1):
                self.emit('page-requested', 'end')
                return

            if first_page.loadable and first_page._ic_position + first_page._ic_height < -self.widget_height * PRELOAD:
                self.remove(first_page)
                return

        if self.scroll_direction in (Gtk.DirectionType.UP, None):
            self.scroll_direction = Gtk.DirectionType.UP

            if first_page.loadable and first_page._ic_position > -self.widget_height * PRELOAD:
                self.emit('page-requested', 'start')
                return

            if last_page.loadable and last_page._ic_position > self.widget_height * (PRELOAD + 1):
                self.remove(last_page)
                return

    def append(self, page):
        """ Adds page at end """
        last_page = self.get_last_child()

        page._ic_height = self.get_height()
        page._ic_position = last_page._ic_position + last_page._ic_height if last_page else 0

        page.insert_before(self, None)

        page.connect('rendered', self.on_page_rendered)
        page.render()

    def cancel_deceleration(self):
        # Assume grandparent of `self` is a Gtk.ScrolledWindow
        self.get_parent().get_parent().set_kinetic_scrolling(False)
        self.get_parent().get_parent().set_kinetic_scrolling(True)

    def clear(self):
        page = self.get_last_child()
        while page:
            prev_page = page.get_prev_sibling()
            page.dispose()
            page = prev_page

        self.scroll_direction = None
        self.queue_allocate()

    def configure_adjustments(self):
        lower = 0
        if first_page := self.get_first_child():
            if first_page.status == 'offlimit' and self.scroll_direction in (None, Gtk.DirectionType.UP):
                # Offlimit start
                lower += self.widget_height
                if first_page._ic_position == -self.widget_height and self.scroll_direction:
                    self.emit('offlimit', 'start')

        upper = self.canvas_height
        if last_page := self.get_last_child():
            if last_page.status == 'offlimit' and self.scroll_direction in (None, Gtk.DirectionType.DOWN):
                # Offlimit end
                upper -= self.widget_height
                if last_page._ic_position == self.widget_height and self.scroll_direction:
                    self.emit('offlimit', 'end')

        self.vadjustment.configure(
            max(min(self.vadjustment.props.value + self.scroll_adjusting_delta, self.max_vadjustment_value), 0),
            lower,
            upper,
            self.widget_height * 0.1,
            self.widget_height * 0.9,
            min(self.widget_height, self.canvas_height)
        )
        self.scroll_adjusting_delta = 0

    def connect_signals(self):
        if self.vadjustment and self.vadjustment_value_changed_handler_id is None:
            self.vadjustment_value_changed_handler_id = self.vadjustment.connect(
                'value-changed', lambda _adj: self.queue_allocate()
            )

        # Keyboard navigation
        self.key_pressed_handler_id = self.pager.window.controller_key.connect(
            'key-pressed', self.on_key_pressed
        )

    def disconnect_signals(self):
        if self.vadjustment_value_changed_handler_id:
            self.vadjustment.disconnect(self.vadjustment_value_changed_handler_id)
            self.vadjustment_value_changed_handler_id = None

        if self.key_pressed_handler_id:
            self.pager.window.controller_key.disconnect(self.key_pressed_handler_id)
            self.key_pressed_handler_id = None

    def dispose(self):
        self.disconnect_signals()
        self.remove_controllers()
        self.clear()

    def do_size_allocate(self, width, height, baseline, adjusted=False):
        # Keep scrolling position when resized horizontally
        if self.prev_width and self.prev_width != width:
            self.vadjustment.props.value = width * self.vadjustment.props.value / self.prev_width
            self.prev_width = width
            self.do_size_allocate(width, height, baseline, adjusted)
            return

        self.prev_width = width
        self.current_page_top = None
        self.current_page_bottom = None

        page = self.get_first_child()
        if page is None:
            # Empty! Probably in destruction.
            return

        size = 0
        scroll_offset = round(self.vadjustment.props.value) + self.scroll_adjusting_delta

        while page:
            page._ic_position = size - scroll_offset

            if page.image and not page.error:
                _, page_height, _, _ = page.image.do_measure(Gtk.Orientation.VERTICAL, width)
            else:
                page_height = height

            page._ic_height = page_height

            if not self.current_page_top and page._ic_position <= 0 and page._ic_position + page._ic_height > 0:
                self.current_page_top = page
            if not self.current_page_bottom and page._ic_position <= height and page._ic_position + page._ic_height > height:
                self.current_page_bottom = page

            visible = page._ic_position >= 0 and page._ic_position < height
            visible |= page._ic_position + page_height >= 0 and page._ic_position + page_height < height
            visible |= page._ic_position < 0 and page._ic_position + page_height >= height
            page.obscured = not visible

            if page.status in ('rendering', 'allocable') and not page.activity_indicator.get_visible():
                page.activity_indicator.set_visible(visible)

            position = Graphene.Point()
            position.init(0, page._ic_position)

            transform = Gsk.Transform.translate(Gsk.Transform.new(), position)
            page.allocate(width, page_height, baseline, transform)

            size += page_height
            page = page.get_next_sibling()

        self.canvas_height = size
        self.configure_adjustments()
        self.is_scroll_adjusting = False

        self.add_or_remove_page()

    def on_gesture_click_released(self, _gesture, n_press, x, y):
        if n_press != 1:
            return

        if x < self.canvas_width / 3:
            # First third: scroll up
            self.scroll_by_increment(-self.vadjustment.props.page_size * self.settings.scroll_click_percentage)
        elif x > 2 * self.canvas_width / 3:
            # Last third: scroll down
            self.scroll_by_increment(self.vadjustment.props.page_size * self.settings.scroll_click_percentage)
        else:
            # Second third: controls zone
            self.emit('controls-zone-clicked')

    def on_gesture_drag_begin(self, _controller, _start_x, _start_y):
        self.scroll_drag_offset = 0

        # If deceleration is not cancelled gestures become buggy!
        # So, no deceleration on touch screen :-(
        self.cancel_deceleration()

    def on_gesture_drag_end(self, _controller, _offset_x, _offset_y):
        self.cancel_deceleration()

    def on_gesture_drag_update(self, _controller, _offset_x, offset_y):
        if abs(offset_y) < 1:
            # Ignore drags that are only clicks
            # Occurs when a `Retry` button is activated
            return

        self.scroll_direction = Gtk.DirectionType.UP if offset_y > 0 else Gtk.DirectionType.DOWN
        self.vadjustment.props.value -= (offset_y - self.scroll_drag_offset) * self.settings.scroll_drag_factor
        self.scroll_drag_offset = offset_y

        self.gesture_drag.set_state(Gtk.EventSequenceState.CLAIMED)

    def on_key_pressed(self, _controller, keyval, _keycode, state):
        if self.pager.window.page != self.pager.reader.props.tag:
            return Gdk.EVENT_PROPAGATE

        modifiers = Gtk.accelerator_get_default_mod_mask()
        if (state & modifiers) not in (Gdk.ModifierType.NO_MODIFIER_MASK, Gdk.ModifierType.SHIFT_MASK):
            return Gdk.EVENT_PROPAGATE

        if keyval in (Gdk.KEY_plus, Gdk.KEY_KP_Add):
            # Zoom in
            self.zoom = self.pager.set_clamp_size(self.zoom * ZOOM_FACTOR_SCROLL_WHEEL)
            return Gdk.EVENT_STOP

        elif keyval in (Gdk.KEY_minus, Gdk.KEY_KP_Subtract):
            # Zoom out
            self.zoom = self.pager.set_clamp_size(self.zoom / ZOOM_FACTOR_SCROLL_WHEEL)
            return Gdk.EVENT_STOP

        elif keyval in (Gdk.KEY_0, Gdk.KEY_KP_0):
            # Zoom reset
            self.zoom = self.pager.set_clamp_size()
            return Gdk.EVENT_STOP

        elif keyval in (Gdk.KEY_Down, Gdk.KEY_KP_Down, Gdk.KEY_Right, Gdk.KEY_KP_Right, Gdk.KEY_space, Gdk.KEY_k):
            self.emit('keyboard-navigation')
            self.scroll_by_increment(self.vadjustment.props.step_increment, 80, False)
            return Gdk.EVENT_STOP

        elif keyval in (Gdk.KEY_Up, Gdk.KEY_KP_Up, Gdk.KEY_Left, Gdk.KEY_KP_Left, Gdk.KEY_j):
            self.emit('keyboard-navigation')
            self.scroll_by_increment(-self.vadjustment.props.step_increment, 80, False)
            return Gdk.EVENT_STOP

        elif keyval in (Gdk.KEY_Page_Down, Gdk.KEY_KP_Page_Down):
            self.emit('keyboard-navigation')
            self.scroll_by_increment(self.vadjustment.props.page_size * self.settings.scroll_click_percentage)
            return Gdk.EVENT_STOP

        elif keyval in (Gdk.KEY_Page_Up, Gdk.KEY_KP_Page_Up):
            self.emit('keyboard-navigation')
            self.scroll_by_increment(-self.vadjustment.props.page_size * self.settings.scroll_click_percentage)
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def on_page_rendered(self, page, update, retry):
        if not update:
            page.activity_indicator.set_visible(False)

            if retry:
                # No idea why this reset is necessary
                self.gesture_drag.reset()

    def on_page_status_changed(self, page, _status, init_height):
        if page.status not in ('allocable', 'offlimit'):
            return

        if page.status == 'allocable':
            # As soon as page height is known
            # Adjust scroll value if page has been prepended and still above scroll position
            if page._ic_position < 0:
                _, page_height, _, _ = page.image.do_measure(Gtk.Orientation.VERTICAL, self.get_width())
                self.scroll_adjusting_delta = page_height - init_height
                self.queue_allocate()
        else:
            # Offlimit: vadjustment lower or upper value must be updated
            self.configure_adjustments()
            self.is_scroll_adjusting = False

    def on_scroll(self, _controller, _dx, dy):
        self.is_scroll_decelerating = False
        self.scroll_direction = Gtk.DirectionType.UP if dy < 0 else Gtk.DirectionType.DOWN

        modifiers = Gtk.accelerator_get_default_mod_mask()
        state = self.controller_scroll.get_current_event_state()
        if state & modifiers == Gdk.ModifierType.CONTROL_MASK:
            if self.scroll_direction == Gtk.DirectionType.UP:
                zoom = self.zoom * ZOOM_FACTOR_SCROLL_WHEEL
            else:
                zoom = self.zoom / ZOOM_FACTOR_SCROLL_WHEEL
            self.zoom = self.pager.set_clamp_size(zoom)

            return Gdk.EVENT_STOP

        # Scroll deltas unit Gdk.ScrollUnit.SURFACE occurs with touchpads under Wayland
        unit_multiple_factor = 1 if self.controller_scroll.get_unit() == Gdk.ScrollUnit.SURFACE else self.vadjustment.props.step_increment
        self.vadjustment.props.value = min(
            self.vadjustment.props.upper - self.vadjustment.props.page_size,
            max(
                0,
                self.vadjustment.props.value + dy * unit_multiple_factor
            )
        )

        return Gdk.EVENT_STOP

    def on_scroll_decelerate(self, _controller, vx, vy):
        self.is_scroll_decelerating = True

    def prepend(self, page):
        """ Adds a new page at start """
        self.is_scroll_adjusting = True

        page._ic_height = self.get_height()
        page._ic_position = self.get_first_child()._ic_position - page._ic_height

        page.insert_before(self, self.get_first_child())
        self.vadjustment.props.value += page._ic_height

        page.connect('rendered', self.on_page_rendered)
        page.connect('notify::status', self.on_page_status_changed, page._ic_height)
        page.render()

    def print(self, action=''):
        print(f'\n==============={action}================')
        count = 0
        page = self.get_first_child()
        while page:
            index = page.index + 1 or '?'
            chapter_title = page.chapter.title if page.chapter else '?'
            print(f'{count + 1:2}: p={int(page._ic_position):5} | h={int(page._ic_height):5} | {index:3} {chapter_title}')  # noqa
            count += 1
            page = page.get_next_sibling()
        print('================================')

    def remove(self, page):
        """ Removes a page """
        page.dispose()

        if self.scroll_direction == Gtk.DirectionType.DOWN:
            self.vadjustment.props.value -= page._ic_height

    def remove_controllers(self):
        self.remove_controller(self.controller_scroll)
        self.remove_controller(self.gesture_click)
        self.pager.remove_controller(self.gesture_drag)

    def scroll_by_increment(self, increment, duration=500, easing=True):
        if self.is_scroll_by_increment or self.is_scroll_adjusting:
            return

        self.is_scroll_decelerating = False
        self.is_scroll_by_increment = True
        self.scroll_direction = Gtk.DirectionType.UP if increment < 0 else Gtk.DirectionType.DOWN

        start = self.vadjustment.props.value
        end = start + increment

        clock = self.get_frame_clock()
        start_time = clock.get_frame_time()
        end_time = start_time + duration * 1000

        def ease_out_cubic(t):
            return (t - 1) ** 3 + 1

        def on_finish():
            self.is_scroll_by_increment = False

            self.queue_allocate()

        def tick_callback(_self, clock):
            now = clock.get_frame_time()

            if now < end_time and self.vadjustment.props.value != end:
                t = (now - start_time) / (end_time - start_time)
                if easing:
                    t = ease_out_cubic(t)
                self.vadjustment.props.value = start + t * (end - start)

                return GLib.SOURCE_CONTINUE

            self.vadjustment.props.value = end

            GLib.idle_add(on_finish)

            return GLib.SOURCE_REMOVE

        self.add_tick_callback(tick_callback)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.manga_stream import MangaStream


class Pointzerotoons(MangaStream):
    id = 'pointzerotoons'
    name = 'Point Zero Toons'
    lang = 'pt_BR'

    date_format = '%d.%m.%Y'

    base_url = 'https://kitsuneyako.com'
    logo_url = base_url + '/wp-content/uploads/2026/01/cropped-Imagem-do-WhatsApp-de-2025-10-18-as-12.15.42_53866798-32x32.jpg'

    authors_selector = '.tsinfo .imptdt:-soup-contains("Artista") i, .tsinfo .imptdt:-soup-contains("Autor") i'
    genres_selector = '.tx-hero-genres a'
    scanlators_selector = '.tsinfo .imptdt:-soup-contains("Postado Por") i'
    status_selector = '.tsinfo .imptdt:-soup-contains("Status") i'
    synopsis_selector = '[itemprop="description"]'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.wpcomics import WPComics


class Readcomicfree(WPComics):
    id = 'readcomicfree'
    name = 'Read Comics Free'
    lang = 'en'
    is_nsfw = True
    status = 'disabled'

    base_url = 'https://readcomicfree.org'
    search_url = base_url + '/search'
    latest_updates_url = base_url + '/comic-updates'
    most_populars_url = base_url + '/popular-comics'
    manga_url = base_url + '/comic/{0}'
    chapter_url = base_url + '/comic/{0}/{1}/all'

    details_name_selector = '.breadcrumb li:last-child span'
    details_cover_selector = '#item-detail .col-image img'
    details_status_selector = '#item-detail .list-info .status p:last-child'
    details_authors_selector = '#item-detail .list-info .author p:last-child a'
    details_genres_selector = '#item-detail .list-info .kind p:last-child a'
    details_synopsis_selector = '#item-detail .detail-content p'
    results_link_selector = 'h3 a'
    results_cover_img_selector = '.box_img a img'
    results_last_chapter_link_selector = 'figcaption ul li a'
    results_last_chapter_lastest_updates_link_selector = '.hlb-list li a'



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _

from bs4 import BeautifulSoup

from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.webview import CompleteChallenge


class Readcomicsonline(Server):
    id = 'readcomicsonline'
    name = 'Read Comics Online'
    lang = 'en'

    has_cf = True

    base_url = 'https://readcomicsonline.ru'
    search_url = base_url + '/advanced-search'
    manga_list_url = base_url + '/comic-list'
    manga_url = base_url + '/comic/{0}'
    chapter_url = base_url + '/comic/{0}/{1}'
    image_url = 'https://cdn.readcomicsonline.ru/uploads/manga/{0}/chapters/{1}/{2}'

    filters = [
        {
            'key': 'status',
            'type': 'select',
            'name': _('Status'),
            'description': _('Filter by Status'),
            'value_type': 'single',
            'default': '',
            'options': [
                {'key': '', 'name': _('All')},
                {'key': '1', 'name': _('Ongoing')},
                {'key': '2', 'name': _('Complete')},
            ],
        },
        {
            'key': 'type',
            'type': 'select',
            'name': _('Type'),
            'description': _('Filter by Type'),
            'value_type': 'single',
            'default': '',
            'options': [
                {'key': '', 'name': _('All')},
                {'key': '1', 'name': 'DC Comics'},
                {'key': '2', 'name': 'Marvel Comics'},
                {'key': '3', 'name': _('Other')},
            ],
        },
    ]

    def __init__(self):
        self.session = None

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns comic data by scraping comic HTML page content

        Initial data should contain at least comic's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Comic slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update({
            'authors': [],
            'scanlators': [],  # not available
            'genres': [],
            'status': None,
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
            'cover': None,
        })

        data['name'] = soup.select_one('main h1').text.strip()
        data['cover'] = soup.select_one('main img.object-cover').get('src')

        # Details
        if element := soup.select_one('main span.rounded-full'):
            status = element.text.strip()
            if status == 'Complete':
                data['status'] = 'complete'
            elif status == 'Ongoing':
                data['status'] = 'ongoing'

        for element in soup.select('div:-soup-contains("Genres") > a'):
            data['genres'].append(element.text.strip())

        for element in soup.select('div:-soup-contains("Author") > a'):
            data['authors'].append(element.text.strip())

        if element := soup.select_one('main p.leading-relaxed'):
            data['synopsis'] = element.text.strip()

        # Chapters
        for element in reversed(soup.select('section a')):
            slug = element.get('href').split('/')[-1]
            title = element.select_one('span:first-child > span').text.strip()
            date = element.select_one('span:last-child').text.strip()

            data['chapters'].append({
                'slug': slug,
                'title': title,
                'date': convert_date_string(date, languages=[self.lang], format='%d %b %Y'),
            })

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns comic chapter data by scraping chapter HTML page content

        Currently, only pages (list of images filenames) are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = {
            'pages': [],
        }
        for element in soup.select('#reader-all img'):
            slug = element.get('src').split('/')[-1]

            data['pages'].append({
                'slug': slug,
                'image': None,
            })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            self.image_url.format(manga_slug, chapter_slug, page['slug']),
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': page['slug'],
        }

    def get_manga_url(self, slug, url):
        """
        Returns comic absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, term=None, orderby=None):
        r = self.session_get(
            self.manga_list_url,
            headers={
                'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
                'Referer': self.manga_list_url,
            },
            params={
                'sort': orderby,
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.grid .group'):
            a_element = element.select_one('a.line-clamp-2')

            results.append({
                'name': a_element.text.strip(),
                'slug': a_element.get('href').split('/')[-1],
                'cover': element.select_one('img.object-cover').get('src'),
                'last_chapter': element.select_one('a.text-xs').text.strip(),
            })

        return results

    @CompleteChallenge()
    def get_latest_updates(self, status=None, type=None):
        """
        Returns list of latest updated comics
        """
        return self.get_manga_list(orderby='latest')

    @CompleteChallenge()
    def get_most_populars(self, status=None, type=None):
        """
        Returns list of most viewed comics
        """
        return self.get_manga_list(orderby='views')

    @CompleteChallenge()
    def search(self, term, status=None, type=None):
        r = self.session_get(self.search_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        if input := soup.select_one('input[name="_token"]'):
            token = input.get('value')
        else:
            return None

        r = self.session_post(
            self.search_url,
            params={
                '_token': token,
                'name':	term,
                'status_id': status if status is not None else '',
                'type_id': type if type is not None else '',
                'category':	'',
            }
        )

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('.grid a.group'):
            slug = element.get('href').split('/')[-1]

            results.append({
                'name': element.p.text.strip(),
                'slug': slug,
                'cover': element.select_one('.rc-cover img').get('src'),
            })

        return results



# SPDX-FileCopyrightText: 2020-2026 GrownNed
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: GrownNed <grownned@gmail.com>
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from datetime import datetime

import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type

SERVER_NAME = 'Desu'

headers = {
    'User-Agent': USER_AGENT,
}


class Desu(Server):
    id = 'desu'
    name = SERVER_NAME
    lang = 'ru'
    is_nsfw = True

    base_url = 'https://desu.uno'
    logo_url = base_url + '/styles/favicons/favicon-32x32.png?v=1'
    api_url = base_url + '/api'
    api_search_url = api_url + '/manga'
    api_manga_url = api_url + '/manga/{0}'
    api_chapters_url = api_url + '/manga/{0}/chapters'
    api_chapter_url = api_url + '/manga/{0}/chapters/{1}'

    manga_title_css_selector = 'h1 > span.name'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = headers

    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(
            self.api_manga_url.format(initial_data['slug']),
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()['manga']

        data = initial_data.copy()
        data.update({
            'authors': [],
            'scanlators': [],
            'genres': [],
            'status': None,
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
        })

        data['name'] = resp_data['russian']
        data['url'] = resp_data['view_url']
        data['cover'] = resp_data['cover']['preview']

        if resp_data.get('translators'):
            data['scanlators'] = [t['name'] for t in resp_data['translators']]
        data['genres'] = [genre['name'] for genre in resp_data['genres']]
        if resp_data['status'] == 'ongoing':
            data['status'] = 'ongoing'
        elif resp_data['status'] == 'released':
            data['status'] = 'complete'
        data['synopsis'] = resp_data['description']

        # Chapters
        r = self.session_get(
            self.api_chapters_url.format(initial_data['slug']),
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()['chapters']

        for chapter in reversed(resp_data):
            title = ''
            if volume := chapter.get('volume'):
                title += f'Том {volume}'
            if number := chapter.get('number'):
                title += f' Глава {number}'
            if chapter.get('title'):
                title += f' - {chapter["title"]}'

            data['chapters'].append({
                'slug': chapter['chapter_id'],
                'title': title,
                'num': number,
                'date': datetime.fromtimestamp(chapter['publish_date']).date(),
            })

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data from API

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.api_chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()['chapter']

        data = {
            'pages': [],
        }
        for page in resp_data['pages']:
            data['pages'].append({
                'slug': None,
                'image': page['url'],
            })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': page['image'].split('/')[-1].split('?')[0],
        }

    @staticmethod
    def get_manga_url(slug, url):
        """
        Returns manga absolute URL
        """
        return url

    def get_latest_updates(self):
        """
        Returns latest updated mangas
        """
        return self.search('', orderby='updated')

    def get_most_populars(self):
        """
        Returns most popular mangas (bayesian rating)
        """
        return self.search('', orderby='popular')

    def search(self, term, orderby=None):
        params = {
            'limit': 50,
        }
        if orderby is not None:
            params['order_by'] = orderby
        else:
            params['search'] = term

        r = self.session_get(
            f'{self.api_search_url}/',
            params=params,
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()['mangas']

        return [{
            'slug': item['manga_id'],
            'name': item['russian'],
            'cover': item['cover']['preview'],
            # last_chapter=item['chapters']['updated']['number'],
        } for item in resp_data]



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.paprika import Paprika


class Mangareadercc(Paprika):
    id = 'mangareadercc'
    name = 'MangaReader (cc)'
    lang = 'en'

    has_cf = True
    is_nsfw = True

    base_url = 'https://mangareader.in'
    logo_url = base_url + '/frontend/imgs/favico.png'
    bypass_cf_url = base_url + '/manga/selena'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.keyoapp import Keyoapp


class Edscanlation(Keyoapp):
    id = 'edscanlation'
    name = 'ED Scanlation'
    lang = 'fr'
    has_cf = True
    status = 'disabled'

    base_url = 'https://edscanlation.fr'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import json
from collections import OrderedDict

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type

SERVER_NAME = 'Manga Eden'

headers = OrderedDict(
    [
        ('User-Agent', USER_AGENT),
        ('Accept-Language', 'en-US,en;q=0.5'),
    ]
)


class Mangaeden(Server):
    id = 'mangaeden'
    name = SERVER_NAME
    lang = 'en'
    status = 'disabled'

    base_url = 'https://www.mangaeden.com'
    search_url = base_url + '/en/en-directory/'
    most_populars_url = search_url + '?order=1'
    manga_url = base_url + '/en/en-manga/{0}/'
    chapter_url = base_url + '/en/en-manga/{0}/{1}/1/'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = headers

    @classmethod
    def get_manga_initial_data_from_url(cls, url):
        return dict(slug=url.split('/')[-2])

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r is None:
            return None

        mime_type = get_buffer_mime_type(r.content)

        if r.status_code != 200 or mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = soup.find('span', class_='manga-title').text.strip()

        cover_element = soup.find('div', class_='mangaImage2')
        if cover_element:
            data['cover'] = 'https:{0}'.format(cover_element.img.get('src'))

        # Details
        for element in soup.find_all('div', class_='rightBox')[1].find_all():
            if element.name == 'h4':
                label = element.text.strip()

                if label.startswith(('Status', 'Stato')):
                    status = element.find_all_next(string=True, limit=2)[1].strip().lower()

                    if status in ('ongoing', 'in corso'):
                        data['status'] = 'ongoing'
                    elif status in ('completed', 'completato'):
                        data['status'] = 'complete'
                    elif status in ('suspended', 'sospeso'):
                        data['status'] = 'suspended'

                continue

            if element.name == 'a':
                if label.startswith(('Author', 'Autore', 'Artist', 'Artista')):
                    data['authors'].append(element.text.strip())
                elif label.startswith(('Genres', 'Genere')):
                    data['genres'].append(element.text.strip())

        # Synopsis
        synopsis_element = soup.find('h2', id='mangaDescription')
        if synopsis_element:
            data['synopsis'] = synopsis_element.text.strip()

        # Chapters
        elements = soup.find('table').tbody.find_all('tr')
        for element in reversed(elements):
            tds_elements = element.find_all('td')

            data['chapters'].append(dict(
                slug=tds_elements[0].a.get('href').split('/')[-3],
                title=tds_elements[0].b.text.strip(),
                date=convert_date_string(tds_elements[3].text.strip(), format='%b %d, %Y'),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r is None:
            return None

        mime_type = get_buffer_mime_type(r.content)

        if r.status_code != 200 or mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for script_element in soup.find_all('script'):
            script = script_element.string
            if not script or not script.strip().startswith('var pages'):
                continue

            pages = json.loads(script.strip().split('\n')[0].split('=')[1][:-1])
            for page in pages:
                data['pages'].append(dict(
                    slug=None,  # not necessary, we know image url already
                    image='https:{0}'.format(page['fs']),
                ))
            break

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'], headers=dict(referer=self.chapter_url.format(manga_slug, chapter_slug)))
        if r is None or r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        """
        Returns most viewed manga list
        """
        r = self.session_get(self.most_populars_url)
        if r is None:
            return None

        mime_type = get_buffer_mime_type(r.content)

        if r.status_code != 200 or mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for tr_element in soup.find('table', id='mangaList').tbody.find_all('tr'):
            td_elements = tr_element.find_all('td')
            a_element = td_elements[0].a
            results.append(dict(
                slug=a_element.get('href').split('/')[-2],
                name=a_element.text.strip(),
            ))

        return results

    def search(self, term):
        r = self.session_get(self.search_url, params=dict(title=term))
        if r is None:
            return None

        mime_type = get_buffer_mime_type(r.content)

        if r.status_code != 200 or mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for tr_element in soup.find('table', id='mangaList').tbody.find_all('tr'):
            td_elements = tr_element.find_all('td')

            if td_elements[3].text.strip() == '0':
                # Skipped manga with no chapters
                continue

            a_element = td_elements[0].a
            results.append(dict(
                slug=a_element.get('href').split('/')[-2],
                name=a_element.text.strip(),
            ))

        return results


class Mangaeden_it(Mangaeden):
    id = 'mangaeden_it'
    name = SERVER_NAME
    lang = 'it'
    status = 'disabled'

    base_url = 'http://www.mangaeden.com'
    search_url = base_url + '/it/it-directory/'
    most_populars_url = search_url + '?order=1'
    manga_url = base_url + '/it/it-manga/{0}/'
    chapter_url = base_url + '/it/it-manga/{0}/{1}/1/'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import json

from bs4 import BeautifulSoup

from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number
from komikku.webview import CompleteChallenge

LANGUAGES_CODES = dict(
    ar='ar',
    cs='cs',
    de='de',
    en='en',
    es='es',
    es_419='es-419',
    fa='fa',
    fr='fr',
    id='id',
    it='it',
    ja='jp',
    ko='kr',
    nl='nl',
    pl='pl',
    pt='pt-pt',
    pt_BR='pt-br',
    ru='ru',
    th='th',
    tr='tr',
    uk='uk',
    vi='vi',
    zh_Hant='zh',
    zh_Hans='zh-cn',
)


class Mangaball(Server):
    id = 'mangaball'
    name = 'MangaBall'
    lang = 'en'

    has_cf = True
    is_nsfw = True

    base_url = 'https://mangaball.net'
    logo_url = base_url + '/public/frontend/images/favicon/favicon-32x32.png'

    search_url = base_url + '/search-advanced/'
    manga_url = base_url + '/title-detail/{0}/'
    chapter_url = base_url + '/chapter-detail/{0}/'

    api_url = base_url + '/api/v1'
    api_search_url = api_url + '/title/search-advanced/'
    api_chapters_url = api_url + '/chapter/chapter-listing-by-title-id/'

    filters = [
        {
            'key': 'demographic',
            'type': 'select',
            'name': _('Demographic'),
            'description': _('Filter by Publication Demographic'),
            'value_type': 'single',
            'default': 'any',
            'options': [
                {'key': 'any', 'name': _('Any')},
                {'key': 'shounen', 'name': _('Shounen')},
                {'key': 'shoujo', 'name': _('Shoujo')},
                {'key': 'seinen', 'name': _('Seinen')},
                {'key': 'josei', 'name': _('Josei')},
                {'key': 'yuri', 'name': _('Yuri')},
                {'key': 'yaoi', 'name': _('Yaoi')},
            ],
        },
        {
            'key': 'status',
            'type': 'select',
            'name': _('Status'),
            'description': _('Filter by Status'),
            'value_type': 'single',
            'default': 'any',
            'options': [
                {'key': 'any', 'name': _('Any')},
                {'key': 'ongoing', 'name': _('Ongoing')},
                {'key': 'completed', 'name': _('Completed')},
                {'key': 'hiatus', 'name': _('Hiatus')},
                {'key': 'cancelled', 'name': _('Canceled')},
            ]
        },
    ]

    def __init__(self):
        self.session = None

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data from manga HTML page

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()

        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        csrf_token = soup.select_one('meta[name="csrf-token"]')['content']

        data['name'] = soup.select_one('#comicDetail h6').text.strip()
        data['cover'] = soup.select_one('.featured-cover').get('src')

        if element := soup.select_one('.badge-status'):
            status = element.text.strip()
            if status == 'Completed':
                data['status'] = 'complete'
            elif status == 'Ongoing':
                data['status'] = 'ongoing'
            elif status == 'Hiatus':
                data['status'] = 'hiatus'
            elif status == 'Cancelled':
                data['status'] = 'suspended'

        for element in soup.select('[data-tag-id]'):
            data['genres'].append(element.text.strip())

        for element in soup.select('[data-person-id]'):
            data['authors'].append(element.text.strip())

        if element := soup.select_one('.description-text > p'):
            data['synopsis'] = element.text.strip()

        # Chapters
        title_id = data['slug'].split('-')[-1]
        r = self.session_post(
            self.api_chapters_url,
            data={
                'title_id': title_id,
                'lang': LANGUAGES_CODES[self.lang],
            },
            headers={
                'Referer': self.manga_url.format(data['slug']),
                'X-Csrf-TOKEN': csrf_token,
                'X-Requested-With': 'XMLHttpRequest',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        for chapter in reversed(resp_data['ALL_CHAPTERS']):
            num = chapter['number_float']
            for translation in chapter['translations']:
                num_volume = translation.get('volume')
                group = translation.get('group')

                data['chapters'].append(dict(
                    slug=translation['id'],
                    title=translation['name'].strip(),
                    scanlators=[group['name']] if group else None,
                    num=num if is_number(num) else None,
                    num_volume=num_volume if is_number(num_volume) else None,
                    date=convert_date_string(translation['date'].split(' ')[0], format='%Y-%m-%d'),
                ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data from chapter HTML page

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for script_element in soup.select('script'):
            script = script_element.string
            if not script or 'const chapterImages' not in script:
                continue

            for line in script.split('\n'):
                line = line.strip()
                if not line.startswith('const chapterImages'):
                    continue

                try:
                    # const chapterImages = JSON.parse(`[...]`);
                    urls = json.loads(line[34:-3])
                    for url in urls:
                        data['pages'].append(dict(
                            slug=None,
                            image=url,
                        ))
                except Exception:
                    return None

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, term=None, demographic=None, status=None, orderby=None):
        # CRSF token
        r = self.session_get(self.search_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        csrf_token = soup.select_one('meta[name="csrf-token"]')['content']

        payload = {}
        if term:
            payload['search_input'] = term

        payload.update({
            'filters[sort]': orderby or 'name_asc',
            'filters[page]': '1',
            'filters[tag_included_mode]': 'and',
            'filters[tag_excluded_mode]': 'and',
            'filters[contentRating]': 'any',
            'filters[demographic]': demographic,
            'filters[person]': 'any',
            'filters[publicationYear]': '',
            'filters[publicationStatus]': status,
            'filters[translatedLanguage][]': LANGUAGES_CODES[self.lang],
        })

        r = self.session_post(
            self.api_search_url,
            data=payload,
            headers={
                'Referer': self.base_url,
                'X-Csrf-TOKEN': csrf_token,
                'X-Requested-With': 'XMLHttpRequest',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        results = []
        for item in resp_data['data']:
            results.append(dict(
                slug=item['url'].split('/')[-2],
                name=item['name'],
                cover=item['cover'],
            ))

        return results

    @CompleteChallenge()
    def get_latest_updates(self, demographic='any', status='any'):
        return self.get_manga_list(demographic=demographic, status=status, orderby='updated_chapters_desc')

    @CompleteChallenge()
    def get_most_populars(self, demographic='any', status='any'):
        return self.get_manga_list(demographic=demographic, status=status, orderby='views_desc')

    @CompleteChallenge()
    def search(self, term, demographic='any', status='any'):
        return self.get_manga_list(term=term, demographic=demographic, status=status)


class Mangaball_ar(Mangaball):
    id = 'mangaball_ar'
    lang = 'ar'


class Mangaball_cs(Mangaball):
    id = 'mangaball_cs'
    lang = 'cs'


class Mangaball_de(Mangaball):
    id = 'mangaball_de'
    lang = 'de'


class Mangaball_es(Mangaball):
    id = 'mangaball_es'
    lang = 'es'


class Mangaball_es_419(Mangaball):
    id = 'mangaball_es_419'
    lang = 'es_419'


class Mangaball_fa(Mangaball):
    id = 'mangaball_fa'
    lang = 'fa'


class Mangaball_fr(Mangaball):
    id = 'mangaball_fr'
    lang = 'fr'


class Mangaball_id(Mangaball):
    id = 'mangaball_id'
    lang = 'id'


class Mangaball_it(Mangaball):
    id = 'mangaball_it'
    lang = 'it'


class Mangaball_ja(Mangaball):
    id = 'mangaball_ja'
    lang = 'ja'


class Mangaball_ko(Mangaball):
    id = 'mangaball_ko'
    lang = 'ko'


class Mangaball_nl(Mangaball):
    id = 'mangaball_nl'
    lang = 'nl'


class Mangaball_pl(Mangaball):
    id = 'mangaball_pl'
    lang = 'pl'


class Mangaball_pt(Mangaball):
    id = 'mangaball_pt'
    lang = 'pt'


class Mangaball_pt_br(Mangaball):
    id = 'mangaball_pt_br'
    lang = 'pt_BR'


class Mangaball_ru(Mangaball):
    id = 'mangaball_ru'
    lang = 'ru'


class Mangaball_th(Mangaball):
    id = 'mangaball_th'
    lang = 'th'


class Mangaball_tr(Mangaball):
    id = 'mangaball_tr'
    lang = 'tr'


class Mangaball_uk(Mangaball):
    id = 'mangaball_uk'
    lang = 'uk'


class Mangaball_vi(Mangaball):
    id = 'mangaball_vi'
    lang = 'vi'


class Mangaball_zh_hant(Mangaball):
    id = 'mangaball_zh_hant'
    lang = 'zh_Hant'


class Mangaball_zh_hans(Mangaball):
    id = 'mangaball_zh_hans'
    lang = 'zh_Hans'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from enum import IntEnum
import logging

from komikku.models.database import create_db_connection
from komikku.models.database import insert_row
from komikku.models.database import update_row

logger = logging.getLogger(__name__)


class Category:
    def __init__(self, row=None):
        if row is not None:
            for key in row.keys():
                setattr(self, key, row[key])

    @classmethod
    def get(cls, id_, db_conn=None):
        if db_conn is None:
            db_conn = create_db_connection()

        row = db_conn.execute('SELECT * FROM categories WHERE id = ?', (id_,)).fetchone()

        if row is None:
            return None

        return cls(row)

    @classmethod
    def new(cls, label):
        data = dict(
            label=label,
        )

        db_conn = create_db_connection()
        with db_conn:
            id_ = insert_row(db_conn, 'categories', data)

        category = cls.get(id_, db_conn=db_conn) if id_ is not None else None

        return category

    @property
    def mangas(self):
        db_conn = create_db_connection()
        rows = db_conn.execute('SELECT manga_id FROM categories_mangas_association WHERE category_id = ?', (self.id,)).fetchall()

        return [row['manga_id'] for row in rows] if rows else []

    def delete(self):
        db_conn = create_db_connection()

        with db_conn:
            db_conn.execute('DELETE FROM categories WHERE id = ?', (self.id, ))

    def update(self, data):
        """
        Updates specific fields

        :param dict data: fields to update
        :return: True on success False otherwise
        """
        ret = False

        for key in data:
            setattr(self, key, data[key])

        db_conn = create_db_connection()
        with db_conn:
            ret = update_row(db_conn, 'categories', self.id, data)

        return ret


class CategoryVirtual(IntEnum):
    ALL = 0
    UNCATEGORIZED = -1



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from abc import ABC
from abc import abstractmethod
import hashlib
from io import BytesIO
import json
import logging
import os
import pickle
import zipfile

from bs4 import BeautifulSoup
try:
    from curl_cffi import requests as crequests
except Exception:
    crequests = None
from http.cookiejar import CookieJar
import requests

from komikku.consts import LOGO_SIZE
from komikku.consts import REQUESTS_TIMEOUT
from komikku.consts import USER_AGENT
from komikku.models.keyring import KeyringHelper
from komikku.models.settings import Settings
from komikku.servers.exceptions import ServerException
from komikku.servers.loader import ServerFinder
from komikku.servers.loader import ServerFinderPriority
from komikku.servers.utils import get_server_main_id_by_id
from komikku.utils import BaseServer
from komikku.utils import get_cache_dir
from komikku.utils import get_cached_logos_dir
from komikku.utils import retry_session

APP_MIN_VERSION = '1.105.0'  # Minimum app version required to use `Up-to-date servers modules`

# https://www.localeplanet.com/icu/
LANGUAGES = dict(
    ar='العربية',
    cs='Čeština',
    de='Deutsch',
    en='English',
    eo='Espéranto',
    es='Español',
    es_419='Español (Latinoamérica)',
    fa='فارسی',
    fr='Français',
    id='Bahasa Indonesia',
    it='Italiano',
    ja='日本語',
    ko='한국어',
    nb='Norsk Bokmål',
    nl='Nederlands',
    pl='Polski',
    pt='Português',
    pt_BR='Português (Brasil)',
    ru='Русский',
    th='ไทย',
    tr='Türkçe',
    uk='Українська',
    vi='Tiếng Việt',
    zh_Hans='中文 (简体)',
    zh_Hant='中文 (繁體)',
)

VERSION = 1

logger = logging.getLogger('komikku.servers')


class Server(BaseServer, ABC):
    id: str
    name: str
    lang: str
    content = None
    true_search = True  # If False, hide search in Explorer search page (XKCD, DBM, pepper&carotte…)

    base_url = None
    donate_url = None
    logo_url = None

    bypass_cf_url = None  # To be used when challenge is not present at base_url
    filters = None  # List of search filters (genre, rating, status…)
    has_captcha = False
    has_cf = False
    has_login = False
    is_nsfw = False
    is_nsfw_only = False
    logged_in = False
    long_strip_genres = []
    manga_title_css_selector = None  # Used to extract manga title in a manga URL
    params = None  # List of parameters (images quality for ex.)
    sync = False

    @classmethod
    def get_manga_initial_data_from_url(cls, url):
        if cls.manga_title_css_selector:
            c = cls()
            r = c.session_get(url)
            if r.status_code != 200:
                return None

            soup = BeautifulSoup(r.text, 'html.parser')

            title_element = soup.select_one(cls.manga_title_css_selector)
            if not title_element:
                return None

            results = c.search(title_element.text.strip())
            if not results:
                return None

            slug = results[0]['slug']
        else:
            slug = url.split('?')[0].split('/')[-1]

        return dict(slug=slug)

    @property
    def logo_path(self):
        path = os.path.join(get_cached_logos_dir(), 'servers', f'{get_server_main_id_by_id(self.id)}.png')
        if not os.path.exists(path):
            return None

        return path

    def clear_session(self, all=False):
        main_id = get_server_main_id_by_id(self.id)

        # Remove session from disk
        file_path = os.path.join(self.sessions_dir, '{0}.pickle'.format(main_id))
        if os.path.exists(file_path):
            os.unlink(file_path)

        if all:
            for id_ in self._BaseServer__sessions.keys():
                if id_.startswith(main_id):
                    del self._BaseServer__sessions[id_]
        elif self.id in self._BaseServer__sessions:
            del self._BaseServer__sessions[self.id]

    def do_login(self, username=None, password=None):
        if username and password:
            # Username and password are provided only when user defines the credentials in the settings
            self.clear_session()

        credential = self.get_credential()
        if self.base_url is None and credential:
            self.base_url = credential.address

        if self.base_url is None:
            logger.warning('Server %s base_url is not defined. Please check server address in Settings', self.name)
            return

        if self.session is None:
            if self.load_session():
                self.logged_in = True
            else:
                self.session = requests.Session()
                if self.headers:
                    self.session.headers = self.headers

                if username is None and password is None:
                    if credential.username and credential.password:
                        self.logged_in = self.login(credential.username, credential.password)
                else:
                    self.logged_in = self.login(username, password)
        else:
            self.logged_in = True

    def get_credential(self):
        return KeyringHelper().get(get_server_main_id_by_id(self.id))

    @abstractmethod
    def get_manga_data(self, initial_data):
        """This method must return a dictionary.

        Data are usually obtained:
        - by scrapping an HTML page
        - or by parsing the response of a request to an API.

        In most cases, the URL of the HTML page or the URL of the API endpoint
        are forged using a slug provided by method `search` and available in `initial_data` argument.

        By convention, returned dict MUST contain the following keys:
        - name: Name of the manga
        - authors: List of authors (str) [optional]
        - scanlators: List of scanlators (str) [optional]
        - genres: List of genres (str) [optional]
        - status: Status of the manga (See database.Manga.STATUSES) [optional]
        - synopsis: Synopsis of the manga [optional]
        - chapters: List of chapters (See description below)
        - server_id: The server ID
        - cover: Absolute URL of the cover

        By convention, a chapter is a dictionary which MUST contain the following keys:
        - slug: A slug (str) allowing to forge HTML page URL of the chapter
          (usually in conjunction with the manga slug)
        - url: URL of chapter HTML page if `slug` is not usable
        - title: Title of the chapter
        - date: Publish date of the chapter [optional]
        - scanlators: List of scanlators (str) [optional]

        .. warning:: the slug passed to `initial_data` param must never be modified.
        It must remain the same as the one obtained from search,
        because the slug's uniqueness is verified during search.
        If slug is modified, an UNIQUE constraint on (server_id, slug) can be violated when
        inserting manga into the database.
        However, there is one exception: when updating and a change in slug format is detected.
        """

    @abstractmethod
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """This method must return a list of pages.

        Data are usually obtained:
        - by scrapping an HTML page
        - or by parsing the response of a request to an API.

        The URL of the HTML page or the URL of the API endpoint are forged using 4 provided arguments.

        By convention, each page is a dictionary which MUST contain one of the 3 keys `slug`, `image` or `url`:
        - slug : A slug (str) allowing to forge image URL of the page
                 (usually in conjunction with the manga slug and the chapter slug)
        - image: Absolute or relative URL of the page image
        - url: URL of the HTML page to scrape to get the URL of the page image

        It's of course possible to add any other information if necessary
        (an index for example to compute a better image filename).

        The page data are passed to `get_manga_chapter_page_image` method.
        """

    @abstractmethod
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """This method must return a dictionary with the following keys:

        - buffer: Image buffer
        - mime_type: Image MIME type
        - name: Filename of the image

        Depending on the server, we have:
        - the slug or the URL of the image
        - or the URL of the HTML page containing the image.

        In the first case, we have the URL (or can forge it) so we can directly retrieve the image with a GET request.

        In the second case, we must first retrieve the URL of the image by scraping the HTML page containing the image.
        """

    @abstractmethod
    def get_manga_url(self, slug, url):
        """This method must return absolute URL of the manga"""

    def get_param(self, key):
        """Returns param value if defined in server settings else returns param default value"""

        # Check param exists
        if self.params is None and self.filters is None:
            logger.error('[%s] Unknown server param key `%s`', self.id, key)
            raise ServerException('Unknown server param key {0}'.format(key))

        param = None
        for param_ in self.params:
            if param_['key'] == key:
                param = param_
                break

        if param is None:
            # Search param in filters
            for filter in self.filters:
                if filter.get('include_in_settings', False) and filter['key'] == key:
                    param = filter
                    break

        if param is None:
            logger.error('[%s] Unknown server param key `%s`', self.id, key)
            raise ServerException('Unknown server param key {0}'.format(key))

        # Search for param in server settings
        if settings := Settings.get_default():
            main_id = get_server_main_id_by_id(self.id)
            if server_settings := settings.servers_settings.get(main_id):
                if 'params' in server_settings and key in server_settings['params']:
                    return server_settings['params'][key]

        # Return default value
        if param['type'] == 'select' and param['value_type'] == 'multiple':
            return [option['key'] for option in param['options'] if option['default']]
        else:
            return param['default']

    def is_long_strip(self, data):
        """
        Returns True if the manga is a long strip, False otherwise.

        The server shall not modify `data` to form the return value.
        """
        if not self.long_strip_genres:
            return False

        for genre in data['genres']:
            if genre in self.long_strip_genres:
                return True

        return False

    def is_up(self):
        if self.id == 'local':
            return True

        if self.base_url is None:
            # base_url is customizable in settings and stored in keyring
            if credential := self.get_credential():
                self.base_url = credential.address
            else:
                return False

        try:
            requests.get(self.base_url, headers={'User-Agent': USER_AGENT})
        except Exception:
            return False

        return True

    def load_session(self):
        """ Load previous session from disk """

        file_path = os.path.join(self.sessions_dir, '{0}.pickle'.format(get_server_main_id_by_id(self.id)))
        if not os.path.exists(file_path):
            return False

        with open(file_path, 'rb') as f:
            unpickled_session = pickle.load(f)
            if self.http_client == 'requests' and isinstance(unpickled_session, requests.sessions.Session):
                # Pickle data is a `requests` session
                self.session = unpickled_session

            elif self.http_client == 'curl_cffi' and isinstance(unpickled_session, dict):
                # Pickle data is `dict` object containing cookies and headers
                cookie_jar = CookieJar()
                for _domain, dcookies in unpickled_session['cookies'].items():
                    for _path, pcookies in dcookies.items():
                        for _name, cookie in pcookies.items():
                            cookie_jar.set_cookie(cookie)

                self.session = crequests.Session(
                    allow_redirects=True,
                    impersonate='chrome',
                    timeout=(REQUESTS_TIMEOUT, REQUESTS_TIMEOUT * 2),
                    cookies=cookie_jar,
                    headers=unpickled_session['headers']
                )

            else:
                return False

        return True

    def login(self, _username, _password):
        return False

    def save_logo(self):
        return self.save_image(
            self.logo_url, os.path.join(get_cached_logos_dir(), 'servers'), get_server_main_id_by_id(self.id),
            LOGO_SIZE, LOGO_SIZE, keep_aspect_ratio=False, format='PNG'
        )

    def save_session(self):
        """ Save session to disk (cache) """

        file_path = os.path.join(self.sessions_dir, '{0}.pickle'.format(get_server_main_id_by_id(self.id)))
        with open(file_path, 'wb') as f:
            if self.http_client == 'requests':
                pickle.dump(self.session, f)

            elif self.http_client == 'curl_cffi':
                pickle.dump({'cookies': self.session.cookies.jar._cookies, 'headers': self.session.headers}, f)

    @abstractmethod
    def search(self, term=None):
        """This method must return a dictionary.

        Data are usually obtained:
        - by scrapping an HTML page
        - or by parsing the response of a request to an API.

        By convention, returned dict MUST contain the following keys:
        - slug: A slug (str) allowing to forge URL of the HTML page of the manga
        - url: URL of manga HTML page if `slug` is not usable
        - name: Name of the manga
        - cover: Absolute URL of the manga cover [optional]
        - last_chapter: last chapter available [optional]
        - nb_chapters: number of chapters available [optional]

        The data are passed to `get_manga_data` method.

        .. note:: It's of course possible to add any other information/keys if necessary
        .. warning:: but these must not be present in the return value of `get_manga_data`.
        """

    def update_chapter_read_progress(self, data, manga_slug, manga_name, chapter_slug, chapter_url):
        return NotImplemented


class ServerDummy(Server):
    """
    Used as fallback in case the server for a comic in the library has been removed
    """

    def __init__(self, id):
        self.id = id
        self.name = id
        self.lang = None

    def get_manga_data(self, initial_data):
        return None

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        return None

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        return None

    def get_manga_url(self, slug, url):
        return None

    def search(self, term=None):
        return None


def init_servers_modules(use_external_modules, reload_modules=False):
    # Add a first Finder with HIGH priority
    server_finder = ServerFinder(priority=ServerFinderPriority.HIGH)
    server_finder.add_path(os.environ.get('KOMIKKU_SERVERS_PATH'))

    if use_external_modules:
        # A single Finder is sufficient
        server_finder.add_path(os.path.join(get_cache_dir(), 'servers/repo'))
        server_finder.install()
    else:
        # Install first Finder
        server_finder.install()

        # Add a second Finder with LOW priority
        # Once installed, external servers modules must still be accessible, as one or more new servers modules
        # (not present in the application servers modules) may potentially have been used
        # (i.e. comics from these servers modules have been added to the library).
        server_finder = ServerFinder(ServerFinderPriority.LOW)
        server_finder.add_path(os.path.join(get_cache_dir(), 'servers/repo'))
        server_finder.install()


def install_servers_modules_from_repo(app_version):
    """
    Installs (or updates) a alternative version of servers modules from the source code repository.

    The intention is to be able to use the most up-to-date versions of servers modules,
    including new servers, without waiting for a new version of the application.

    Can only work if the application version is greater than or equal to the minimum version
    required in the index.json file.
    """
    repo_url = 'https://febvre.info'
    dest_path = os.path.join(get_cache_dir(), 'servers/repo')
    index_path = os.path.join(dest_path, 'index.json')

    session = retry_session()
    session.headers.update({'User-Agent': f'komikku-{app_version}'})

    def install_zip(current_hash=None):
        # Get remote index.json
        url = repo_url + '/komikku/index.json'
        try:
            r = session.get(url)
        except Exception:
            return None, None

        if r.status_code != 200:
            return None, None

        try:
            app_min_version = json.loads(r.content)['app_min_version']
        except Exception:
            return None, None

        if app_min_version and app_version < app_min_version:
            return False, 'forbidden'

        if current_hash:
            remote_hash = hashlib.sha256(r.content).hexdigest()
            if current_hash == remote_hash:
                return False, 'unchanged'

        url = repo_url + '/komikku/servers.zip'
        try:
            r = session.get(url)
        except Exception:
            return None, None

        if r.status_code != 200:
            return None, None

        with BytesIO(r.content) as io_buffer:
            with zipfile.ZipFile(io_buffer) as zip:
                for zip_info in zip.infolist():
                    if zip_info.is_dir():
                        continue

                    zip.extract(zip_info, dest_path)

        return True, 'updated' if current_hash else 'created'

    if not os.path.exists(dest_path) or not os.path.exists(index_path):
        os.makedirs(dest_path, exist_ok=True)

        return install_zip()

    # Compute current index.json hash
    with open(index_path, 'rb') as fp:
        current_hash = hashlib.sha256(fp.read()).hexdigest()

    return install_zip(current_hash)



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.foolslide import FoOlSlide


class Nifteam(FoOlSlide):
    id = 'nifteam'
    name = 'NIF Team'
    lang = 'it'

    base_url = 'https://read-nifteam.info'
    search_url = base_url + '/slide/search/'
    latest_updates_url = base_url + '/slide/'
    manga_list_url = base_url + '/slide/directory/'
    manga_url = base_url + '/slide/series/{0}'
    chapter_url = base_url + '/slide/read/{0}/it/{1}/page/1'

    name_selector = '.title h1'
    details_selector = None
    synopsis_selector = '.comic .info'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import time

from bs4 import BeautifulSoup
import requests

from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import get_soup_element_inner_text
from komikku.utils import get_buffer_mime_type
from komikku.utils import get_response_elapsed
from komikku.utils import is_number

SEARCH_RESULTS_PAGES = 2

# Seems to be linked to MangaHere


class Mangatown(Server):
    id = 'mangatown'
    name = 'MangaTown'
    lang = 'en'

    is_nsfw = True

    base_url = 'https://www.mangatown.com'
    logo_url = 'https://static.mangatown.com/v20251117/mangatown/images/favicon.ico'

    manga_url = base_url + '/title/{0}'
    search_url = base_url + '/search'
    manga_list_url = base_url + '/directory/{0}-{1}-{2}-{3}-{4}-{5}/'
    manga_url = base_url + '/manga/{0}/'
    chapter_url = base_url + '/manga/{0}/{1}/'
    page_url = base_url + '/manga/{0}/{1}/{2}.html'

    filters = [
        {
            'key': 'demographic',
            'type': 'select',
            'name': _('Demographic'),
            'description': _('Filter by Publication Demographic'),
            'value_type': 'single',
            'default': '',
            'options': [
                {'key': '', 'name': _('All')},
                {'key': 'shounen', 'name': _('Shounen')},
                {'key': 'seinen', 'name': _('Seinen')},
                {'key': 'shoujo', 'name': _('Shoujo')},
                {'key': 'yaoi', 'name': _('Yaoi')},
                {'key': 'shoujo_ai', 'name': _('Shoujo Ai')},
                {'key': 'josei', 'name': _('Josei')},
                {'key': 'shounen_ai', 'name': _('Shounen Ai')},
                {'key': 'yuri', 'name': _('Yuri')},
            ]
        },
        {
            'key': 'status',
            'type': 'select',
            'name': _('Status'),
            'description': _('Filter by Status'),
            'value_type': 'single',
            'default': '',
            'options': [
                {'key': '', 'name': _('All')},
                {'key': 'new', 'name': _('New')},
                {'key': 'ongoing', 'name': _('Ongoing')},
                {'key': 'completed', 'name': _('Completed')},
            ]
        },
        {
            'key': 'type',
            'type': 'select',
            'name': _('Type'),
            'description': _('Filter by Type'),
            'value_type': 'single',
            'default': '',
            'options': [
                {'key': '', 'name': _('All')},
                {'key': 'manga', 'name': _('Manga')},
                {'key': 'manhua', 'name': _('Manhua')},
                {'key': 'manhwa', 'name': _('Manhwa')},
            ]
        },
    ]

    headers = {
        'User-Agent': USER_AGENT,
    }

    long_strip_genres = [
        'Webtoons',
    ]

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = self.headers

    def get_manga_data(self, initial_data):
        """
        Returns manga data from manga HTML page

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = soup.select_one('h1.title-top').text.strip()
        data['cover'] = soup.select_one('.detail_info > img').get('src')

        if element := soup.select_one('.detail_info li:-soup-contains("Author") a'):
            data['authors'].append(element.text.strip())
        if element := soup.select_one('.detail_info li:-soup-contains("Artist") a'):
            artist = element.text.strip()
            if artist not in data['authors']:
                data['authors'].append(artist)

        if element := soup.select_one('.detail_info li:-soup-contains("Status")'):
            status = get_soup_element_inner_text(element, recursive=False).split()[0]
            if status == 'Ongoing':
                data['status'] = 'ongoing'
            elif status == 'Completed':
                data['status'] = 'complete'

        for element in soup.select('.detail_info li:-soup-contains("Genre") a'):
            data['genres'].append(element.text.strip())
        if element := soup.select_one('.detail_info li:-soup-contains("Demographic") a'):
            data['genres'].append(element.text.strip())
        if element := soup.select_one('.detail_info li:-soup-contains("Type") a'):
            data['genres'].append(element.text.strip())

        if element := soup.select_one('.detail_info li:-soup-contains("Summary") #show'):
            data['synopsis'] = element.text.replace('HIDE', '').strip()

        for element in reversed(soup.select('.chapter_list li')):
            slug = element.a.get('href').split('/')[-2]
            num = slug[1:]
            date = element.select_one('.time').text.replace('Today', '').replace('Yesterday', '').strip()

            data['chapters'].append({
                'slug': slug,
                'title': f'Ch. {num}',
                'num': num if is_number(num) else None,
                'date': convert_date_string(date, format='%b %d,%Y', languages=[self.lang]),
            })

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data from chapter HTML page
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        if soup.select_one('#viewer #image'):
            # Single page mode
            if element := soup.select_one('.page_select select'):
                for option_element in element.select('option'):
                    url = option_element.get('value')
                    slug = url.split('/')[-1].replace('.html', '')
                    if slug == 'featured':
                        continue
                    if slug == '':
                        slug = '1'

                    data['pages'].append({
                        'slug': slug,
                        'image': None,
                    })
        else:
            # Webtoon mode
            for img_element in soup.select('#viewer .image'):
                url = img_element.get('src')

                data['pages'].append({
                    'slug': None,
                    'image': img_element.get('src'),
                })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        if slug := page['slug']:
            r = self.session_get(self.page_url.format(manga_slug, chapter_slug, slug))
            if r.status_code != 200:
                return None

            soup = BeautifulSoup(r.text, 'lxml')

            if img_element := soup.select_one('#image'):
                url = img_element.get('src')
            else:
                return None
        else:
            url = page['image']

        if not url.startswith('https:'):
            url = f'https:{url}'  # noqa

        r = self.session_get(
            url,
            headers={
                'Referer': self.page_url.format(manga_slug, chapter_slug, page['slug']),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=url.split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, term=None, demographic=None, status=None, type=None, orderby=None):
        def get_page(page):
            params = {}

            if term:
                url = self.search_url

                params['page'] = page
                params['name_method'] = 'cw'
                params['name'] = term

                if type is not None:
                    params['type'] = type
                if demographic is not None:
                    params['demographic'] = demographic.replace('_', '').upper()
                if status is not None:
                    if status in ('new', 'ongoing'):
                        params['is_completed'] = 0
                    elif status == 'completed':
                        params['is_completed'] = 1
            else:
                url = self.manga_list_url
                if page > 1:
                    url = f'{url}{page}.html'
                if orderby == 'latest':
                    url = f'{url}?last_chapter_time.za'

                url = url.format(
                    demographic if demographic else 0,
                    0,  # genre
                    0,  # year
                    status if status else 0,
                    0,  # alphabetic
                    type if type else 0,
                )

            r = self.session_get(
                url,
                params=params,
                headers={
                    'Referer': f'{self.base_url}/',
                }
            )
            if r.status_code != 200:
                return [], False, None

            soup = BeautifulSoup(r.text, 'lxml')

            items = []
            for element in soup.select('.manga_pic_list li'):
                cover_element = element.select_one('.manga_cover img')
                a_element = element.select_one('.title a')
                last_chapter_element = element.select_one('.new_chapter a')

                items.append({
                    'slug': a_element.get('href').split('/')[-2],
                    'name': a_element.get('title'),
                    'cover': cover_element.get('src'),
                    'last_chapter': last_chapter_element.get('title').split()[-1],
                })

            more = 'javascript' not in soup.select_one('.next-page a.next').get('href') and page < SEARCH_RESULTS_PAGES

            return items, more, get_response_elapsed(r)

        results = []
        delay = None
        more = True
        page = 1
        while more:
            if delay:
                time.sleep(delay)

            items, more, rtime = get_page(page)
            results += items

            delay = min(rtime * 4, DOWNLOAD_MAX_DELAY) if rtime else None
            page += 1

        return results

    def get_latest_updates(self, demographic=None, status=None, type=None):
        return self.get_manga_list(demographic=demographic, status=status, type=type, orderby='latest')

    def get_most_populars(self, demographic=None, status=None, type=None):
        return self.get_manga_list(demographic=demographic, status=status, type=type, orderby='popular')

    def search(self, term, demographic=None, status=None, type=None):
        return self.get_manga_list(term=term, demographic=demographic, type=type, status=status)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.peachscan import Peachscan


class Nazarickscan(Peachscan):
    id = 'nazarickscan'
    name = 'Nazarick Scan'
    lang = 'pt_BR'
    status = 'disabled'

    base_url = 'https://nazarickscan.com.br'
    logo_url = base_url + '/static/nazarickscan.com.br/favicon-32x32.png'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import gi
import os
import platform

import magic

gi.require_version('Adw', '1')
gi.require_version('GdkPixbuf', '2.0')
gi.require_version('Gtk', '4.0')
gi.require_version('Soup', '3.0')
gi.require_version('WebKit', '6.0')

from gi.repository import Adw
from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Soup
from gi.repository import WebKit

from komikku.models.database import VERSION as DB_VERSION
from komikku.utils import check_cmdline_tool


class DebugInfo:
    def __init__(self, app):
        self.app = app

    def get_flatpak_info(self):
        path = os.path.join(GLib.get_user_runtime_dir(), 'flatpak-info')
        if not os.path.exists(path):
            return None

        data = [
            ('Application', 'runtime'),
            ('Instance', 'runtime-commit'),
            ('Instance', 'arch'),
            ('Instance', 'flatpak-version'),
            ('Instance', 'devel'),
        ]

        info = dict()
        keyfile = GLib.KeyFile.new()
        keyfile.load_from_file(path, 0)
        for group, key in data:
            try:
                info[key] = keyfile.get_string(group, key)
            except Exception:
                info[key] = 'N/A'

        return info

    def get_gtk_info(self):
        info = {}

        backend = self.app.window.get_display().__class__.__name__
        if backend == 'GdkX11Display':
            info['backend'] = 'X11'
        elif backend == 'GdkWaylandDisplay':
            info['backend'] = 'Wayland'
        elif backend == 'GdkBroadwayDisplay':
            info['backend'] = 'Broadway'
        elif backend == 'GdkMacosDisplay':
            info['backend'] = 'macOS'
        else:
            info['backend'] = backend

        renderer = self.app.window.get_renderer().__class__.__name__
        if renderer == 'VulkanRenderer':
            info['renderer'] = 'Vulkan'
        elif renderer == 'GLRenderer':
            info['renderer'] = 'GL'
        elif renderer == 'NglRenderer':
            info['renderer'] = 'nGL'
        elif renderer == 'CairoRenderer':
            info['renderer'] = 'Cairo'
        else:
            info['renderer'] = renderer

        info['animations'] = Gtk.Settings.get_default().get_property('gtk-enable-animations')
        info['theme'] = Gtk.Settings.get_default().get_property('gtk-theme-name')
        info['icon-theme'] = Gtk.Settings.get_default().get_property('gtk-icon-theme-name')

        return info

    def get_tools_info(self):
        info = {}

        # Unrar (not available in flatpak sandbox)
        status, ret = check_cmdline_tool('unrar')
        info['unrar'] = ret.split('\n')[0] if status else 'N/A'

        # Unar
        status, ret = check_cmdline_tool(['unar', '-v'])
        if not status:
            # Check flatpak location
            status, ret = check_cmdline_tool(['/app/bin/unar', '-v'])
        info['unar'] = ret if status else 'N/A'

        return info

    def generate(self):
        info = 'Komikku:\n'
        info += f'- Version: {self.app.version}\n'
        info += f'- Profile: {self.app.profile}\n'
        info += f'- DB version: {DB_VERSION}\n'
        info += f'- ID: {self.app.application_id}\n'
        info += '\n'

        info += 'Compiled against:\n'
        info += f'- GLib: {GLib.MAJOR_VERSION}.{GLib.MINOR_VERSION}.{GLib.MICRO_VERSION}\n'
        info += f'- GdkPixbuf: {GdkPixbuf.PIXBUF_VERSION}\n'
        info += f'- GTK: {Gtk.MAJOR_VERSION}.{Gtk.MINOR_VERSION}.{Gtk.MICRO_VERSION}\n'
        info += f'- Adwaita: {Adw.VERSION_S}\n'
        info += f'- WebKitGTK: {WebKit.MAJOR_VERSION}.{WebKit.MINOR_VERSION}.{WebKit.MICRO_VERSION}\n'
        info += f'- Soup: {Soup.MAJOR_VERSION}.{Soup.MINOR_VERSION}.{Soup.MICRO_VERSION}\n'
        info += '\n'

        info += 'Running against:\n'
        info += f'- GTK: {Gtk.get_major_version()}.{Gtk.get_minor_version()}.{Gtk.get_micro_version()}\n'
        info += f'- Adwaita: {Adw.get_major_version()}.{Adw.get_minor_version()}.{Adw.get_micro_version()}\n'
        info += f'- WebKitGTK: {WebKit.get_major_version()}.{WebKit.get_minor_version()}.{WebKit.get_micro_version()}\n'
        info += f'- Soup: {Soup.get_major_version()}.{Soup.get_minor_version()}.{Soup.get_micro_version()}\n'
        info += f'- Magic: {magic.version()}\n'
        info += '\n'

        info += 'System:\n'
        info += f'- Name: {GLib.get_os_info("NAME")}\n'
        info += f'- Version: {GLib.get_os_info("VERSION") or "N/A"}\n'
        info += f'- CPU arch: {platform.machine()}\n'
        info += '\n'

        if flatpak_info := self.get_flatpak_info():
            info += 'Flatpak:\n'
            info += f'- Runtime: {flatpak_info["runtime"]}\n'
            info += f'- Runtime commit: {flatpak_info["runtime-commit"]}\n'
            info += f'- Arch: {flatpak_info["arch"]}\n'
            info += f'- Flatpak version: {flatpak_info["flatpak-version"]}\n'
            info += f'- Devel: {flatpak_info["devel"]}\n'
            info += '\n'

        gtk_info = self.get_gtk_info()
        info += 'GTK:\n'
        info += f"- GDK backend: {gtk_info['backend']}\n"
        info += f"- GSK renderer: {gtk_info['renderer']}\n"
        info += f"- Animations: {gtk_info['animations']}\n"
        info += f"- Theme: {gtk_info['theme']}\n"
        info += f"- Icon theme: {gtk_info['icon-theme']}\n"
        info += '\n'

        info += 'Python:\n'
        info += f'- Version: {platform.python_version()}\n'
        info += f"- PyGObject: {'.'.join(str(v) for v in gi.version_info)}\n"
        info += '\n'

        info += 'Environment:\n'
        info += f'- Desktop: {GLib.getenv("XDG_CURRENT_DESKTOP")}\n'
        info += f'- Session: {GLib.getenv("XDG_SESSION_DESKTOP")} ({GLib.getenv("XDG_SESSION_TYPE")})\n'
        info += f'- Language: {GLib.getenv("LANG")}\n'
        info += f'- Running inside Builder: {"Yes" if GLib.getenv("INSIDE_GNOME_BUILDER") else "No"}\n'
        if gtk_debug := GLib.getenv('GTK_DEBUG'):
            info += f'- GTK_DEBUG: {gtk_debug}\n'
        if gtk_theme := GLib.getenv('GTK_THEME'):
            info += f'- GTK_THEME: {gtk_theme}\n'
        if adw_debug_color_scheme := GLib.getenv('ADW_DEBUG_COLOR_SCHEME'):
            info += f'- ADW_DEBUG_COLOR_SCHEME: {adw_debug_color_scheme}\n'
        if adw_debug_high_contrast := GLib.getenv('ADW_DEBUG_HIGH_CONTRAST'):
            info += f'- ADW_DEBUG_HIGH_CONTRAST: {adw_debug_high_contrast}\n'
        if adw_disable_portal := GLib.getenv('ADW_DISABLE_PORTAL'):
            info += f'- ADW_DISABLE_PORTAL: {adw_disable_portal}\n'
        info += '\n'

        tools_info = self.get_tools_info()
        info += 'Command-line tools:\n'
        info += f'- unrar: {tools_info["unrar"]}\n'
        info += f'- unar: {tools_info["unar"]}'

        return info



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import os

from gi.repository import Adw
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from komikku.explorer.common import ExplorerServerRow
from komikku.explorer.common import set_missing_server_logos
from komikku.models import Settings
from komikku.servers import LANGUAGES
from komikku.servers.utils import get_allowed_servers_list
from komikku.utils import get_data_dir


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/explorer_servers.ui')
class ExplorerServersPage(Adw.NavigationPage):
    __gtype_name__ = 'ExplorerServersPage'

    global_search_button = Gtk.Template.Child('global_search_button')
    search_button = Gtk.Template.Child('search_button')

    searchbar = Gtk.Template.Child('searchbar')
    searchbar_separator = Gtk.Template.Child('searchbar_separator')
    searchentry = Gtk.Template.Child('searchentry')
    listbox = Gtk.Template.Child('listbox')
    pinned_listbox = Gtk.Template.Child('pinned_listbox')

    preselection = False

    def __init__(self, parent):
        Adw.NavigationPage.__init__(self)

        self.parent = parent
        self.window = parent.window

        self.servers = None

        self.connect('shown', self.on_shown)

        self.global_search_button.connect('clicked', self.on_global_search_button_clicked)

        self.searchbar.bind_property(
            'search-mode-enabled', self.search_button, 'active',
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE
        )
        self.searchbar.bind_property(
            'search-mode-enabled', self.searchbar_separator, 'visible',
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE
        )
        self.searchbar.connect_entry(self.searchentry)
        self.searchbar.connect('notify::search-mode-enabled', self.on_search_mode_toggled)
        self.searchbar.set_key_capture_widget(self.window)

        self.searchentry.connect('activate', self.on_searchentry_activated)
        self.searchentry.connect('search-changed', self.search)

        self.pinned_listbox.connect('row-activated', self.on_server_clicked)

        self.listbox.connect('row-activated', self.on_server_clicked)
        self.listbox.set_filter_func(self.filter)

    def clear(self):
        # Clear servers list
        row = self.listbox.get_first_child()
        while row:
            next_row = row.get_next_sibling()
            if isinstance(row, ExplorerServerRow):
                row.dispose()
            row = next_row

        self.listbox.remove_all()

    def clear_pinned(self):
        # Clear pinned servers list
        row = self.pinned_listbox.get_first_child()
        while row:
            next_row = row.get_next_sibling()
            if isinstance(row, ExplorerServerRow):
                row.dispose()
            row = next_row

        self.pinned_listbox.remove_all()

    def filter(self, row):
        """
        This function gets one row and has to return:
        - True if the row should be displayed
        - False if the row should not be displayed
        """
        term = self.searchentry.get_text().strip().lower()

        if not hasattr(row, 'server_data'):
            # Languages headers should always be displayed
            return True

        server_name = row.server_data['name']
        server_lang = row.server_data['lang']

        # Search in name and language
        return (
            term in server_name.lower() or
            term in LANGUAGES.get(server_lang, _('Other')).lower() or
            term in server_lang.lower()
        )

    def on_global_search_button_clicked(self, _button):
        self.parent.search_page.show()

    def on_server_clicked(self, _listbox, row):
        server = getattr(row.server_data['module'], row.server_data['class_name'])()
        if self.preselection and hasattr(row, 'manga_data'):
            self.parent.search_page.show_manga_card(row.manga_data, server)
        else:
            self.parent.search_page.show(server)

    def on_search_mode_toggled(self, _searchbar, _gparam):
        if self.searchbar.get_search_mode():
            self.pinned_listbox.set_visible(False)
        elif len(Settings.get_default().pinned_servers):
            self.pinned_listbox.set_visible(True)

    def on_searchentry_activated(self, _entry):
        if not self.searchbar.get_search_mode():
            return

        # Select first search result
        for child_row in self.listbox:
            if not hasattr(child_row, 'server_data') or not self.filter(child_row):
                continue
            self.on_server_clicked(self.listbox, child_row)
            break

    def on_shown(self, _page):
        if self.searchbar.get_search_mode():
            self.searchentry.grab_focus()

        def do_show_banner():
            if self.window.last_navigation_action != 'push':
                return

            if not Settings.get_default().servers_languages:
                self.window.show_banner(
                    _('It appears that you have not yet defined any server languages in Preferences.'), at_bottom=True
                )

        if not Gtk.Settings.get_default().get_property('gtk-enable-animations'):
            # When animations are disabled, popped/pushed events are sent after `shown` event (bug?)
            # Use idle_add to be sure that last `popped` or `pushed` event has been received
            GLib.idle_add(do_show_banner)
        else:
            do_show_banner()

    def open_local_folder(self, _button):
        path = os.path.join(get_data_dir(), 'local')
        Gio.app_info_launch_default_for_uri(f'file://{path}')  # noqa: E231

    def populate(self, servers=None):
        self.clear()

        if not servers:
            self.servers = get_allowed_servers_list(Settings.get_default())
            self.populate_pinned()
            self.preselection = False
        else:
            self.servers = servers
            self.pinned_listbox.set_visible(False)
            self.preselection = True

        last_lang = None
        for server_data in self.servers:
            if server_data['lang'] != last_lang:
                # Add language header
                last_lang = server_data['lang']

                row = Gtk.ListBoxRow(activatable=False)
                row.add_css_class('explorer-server-section-listboxrow')
                label = Gtk.Label(halign=Gtk.Align.CENTER)
                label.add_css_class('subtitle')
                label.set_text(LANGUAGES.get(server_data['lang'], _('Other')).upper())
                row.set_child(label)
                self.listbox.append(row)

            row = ExplorerServerRow(server_data, self)
            self.listbox.append(row)

        set_missing_server_logos(self.listbox)

        if self.preselection and len(self.servers) == 1:
            row = self.listbox.get_first_child().get_next_sibling()
            self.parent.server = getattr(row.server_data['module'], row.server_data['class_name'])()
            self.parent.search_page.show_manga_card(row.manga_data)
        elif self not in self.window.navigationview.get_navigation_stack():
            self.window.navigationview.push(self)

    def populate_pinned(self):
        self.clear_pinned()

        count = 0
        pinned_servers = Settings.get_default().pinned_servers
        for server_data in self.servers:
            if server_data['id'] not in pinned_servers:
                continue

            row = ExplorerServerRow(server_data, self)
            self.pinned_listbox.append(row)
            count += 1

        set_missing_server_logos(self.pinned_listbox)

        if count:
            # Add header
            row = Gtk.ListBoxRow(activatable=False)
            row.add_css_class('explorer-server-section-listboxrow')
            label = Gtk.Label(halign=Gtk.Align.CENTER)
            label.add_css_class('subtitle')
            label.set_text(_('Pinned').upper())
            row.set_child(label)
            self.pinned_listbox.prepend(row)

        self.pinned_listbox.set_visible(count > 0 and not self.searchbar.get_search_mode())

    def search(self, _entry):
        self.listbox.invalidate_filter()

    def show(self, servers):
        self.searchbar.set_search_mode(False)
        self.populate(servers)

    def toggle_server_pinned_state(self, button, row):
        if button.get_active():
            Settings.get_default().add_pinned_server(row.server_data['id'])
        else:
            Settings.get_default().remove_pinned_server(row.server_data['id'])

        if row.get_parent().get_name() == 'pinned_servers':
            for child_row in self.listbox:
                if not hasattr(child_row, 'server_data'):
                    continue

                if child_row.server_data['id'] == row.server_data['id']:
                    child_row.get_first_child().get_last_child().set_active(button.get_active())
                    break

        self.populate_pinned()



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara

# BEWARE: Submanga server is disabled
# Since ~06/2022, submanga.io domain is redirected to https://lectormanga.online (Leomanga server)


class Submanga(Madara):
    id = 'submanga'
    name = 'Submanga'
    lang = 'es'
    status = 'disabled'

    base_url = 'https://submanga.io'
    chapters_url = base_url + '/manga/{0}/ajax/chapters/'
